import time
import traceback

from ahc.Ahc import ComponentModel, Event, GenericMessage, Lock, \
    GenericMessageHeader, GenericMessagePayload, Topology, EventTypes
from enum import Enum
from datetime import datetime
import uuid

from ahc.Routing.HOLSR.mpr import obtain_mpr
from ahc.Routing.HOLSR.utils import RepeatDeltaTimer, Tracing, keys_to_set


class HOLSRMessageTypes(Enum):
    DATA = "holsr.data"
    HELLO = "holsr.hello"
    TC = "holsr.tc"
    HTC = "holsr.htc"
    CIA = "holsr.cia"


class OLSRLinkStatus(Enum):
    Unidirectional = "uni"
    Bidirectional = "bi"


class OLSRMessageHeader(GenericMessageHeader):
    pass


class OLSRMessagePayload(GenericMessagePayload):
    pass


class OLSRComponent(ComponentModel):
    def __init__(self, component_name, component_id, context=None,
                 configurationparameters=None, num_worker_threads=1):
        super().__init__(component_name, component_id, context=context,
                         configurationparameters=configurationparameters,
                         num_worker_threads=num_worker_threads)
        self.lock = Lock()
        # create an address for the node
        self.address = str(uuid.uuid4())
        # for debugging
        Tracing().register_node(self.address, component_id)
        # handlers for each message kind
        self.message_handlers = {
            HOLSRMessageTypes.DATA: lambda p: self.on_data_handler(p),
            HOLSRMessageTypes.HELLO: lambda p: self.on_hello_handler(p),
            HOLSRMessageTypes.TC: lambda p: self.on_tc_handler(p),
        }
        # list of neighbors to which there
        # exists a valid bidirectional link
        self.valid_bidirectional_links = {}
        # the list of neighbors are heard by this node,
        # but a bidirectional link is not yet validated
        self.non_validated_neighbors = set()
        # neighbor table
        self.neighbor_table = {}
        # mpr set
        self.mpr_set = set()
        self.mpr_sequence_number = 0
        # ms set
        self.ms_set = set()
        self.ms_sequence_number = 0
        # sequence number cache for tc messages
        self.tc_seq_cache = {}
        # routing table
        self.routing_table = {}
        # topology table
        self.topology_table = {}
        # start the component timer
        current_time = datetime.now()
        self.last_hello = current_time
        self.last_tc = current_time
        RepeatDeltaTimer().register_function(self.on_time)

    def on_time(self, current_time, _delta):
        if (current_time - self.last_hello).total_seconds() > 0.25:
            self.last_hello = current_time
            self.broadcast_hello()
        if (current_time - self.last_tc).total_seconds() > 0.5:
            self.last_tc = current_time
            self.broadcast_tc()

    def broadcast_hello(self):
        payload = {'address': self.address}
        self.lock.acquire()
        try:
            payload['mpr_set'] = {a for a in self.mpr_set}
            payload['valid_links'] = keys_to_set(self.valid_bidirectional_links)
            payload['non_validated'] = [a for a in self.non_validated_neighbors]

        finally:
            self.lock.release()
        self.broadcast_msg(HOLSRMessageTypes.HELLO, OLSRMessagePayload(payload))

    def broadcast_tc(self):
        payload = {'address': self.address, 'origin': self.address}
        self.lock.acquire()
        try:
            payload['ms_set'] = {a for a in self.ms_set}
            payload['ms_seq_num'] = self.ms_sequence_number
        finally:
            self.lock.release()
        self.broadcast_msg(HOLSRMessageTypes.TC, OLSRMessagePayload(payload))

    def broadcast_msg(self, msg_type, payload):
        msg_from = self.componentname + "-" + str(self.componentinstancenumber)
        for target in Topology().get_neighbors(self.componentinstancenumber):
            hdr = OLSRMessageHeader(msg_type, msg_from, self.componentname + "-" + str(target),
                                    interfaceid=str(self.componentinstancenumber) + "-" + str(target))
            self.send_down(Event(self, EventTypes.MFRT, GenericMessage(hdr, payload)))

    def on_message_from_bottom(self, eventobj: Event):
        hdr = eventobj.eventcontent.header
        # we expect both from and to fields to
        # start with this component's name
        if hdr.messagefrom.split("-")[0] != self.componentname:
            return
        if hdr.messageto.split("-")[0] != self.componentname:
            return
        payload = eventobj.eventcontent.payload.messagepayload
        # dispatch to the expected message handler
        if hdr.messagetype in self.message_handlers:
            self.lock.acquire()
            try:
                self.message_handlers[hdr.messagetype](payload)
            except Exception as e:
                print(f"ERROR: cannot handle msg of type {hdr.messagetype}: {e}\n")
                traceback.print_exc()
            finally:
                self.lock.release()

    def on_hello_handler(self, payload):
        self.on_hello(payload['address'], payload['valid_links'],
                      payload['non_validated'], payload['mpr_set'])

    def on_hello(self, sender_address, valid_links, non_validated, sender_mpr):
        self.update_links(sender_address, valid_links, non_validated)
        self.select_mpr()
        self.update_ms(sender_address, sender_mpr)
        self.calculate_route_table()
        # for debugging
        if self.componentinstancenumber == 0:
            Tracing().update_edges(list(self.topology_table.keys()))

    def update_links(self, sender_address, valid_links, non_validated):
        # if the sender has us in the non-validated list;
        # then we have already sent them a HELLO message.
        # which means the connection is bidirectionally validated.
        # similarly, if our address is already in the sender's
        # valid links list; update the two-hop list with the new list.
        if self.address in non_validated or self.address in valid_links:
            # promote sender_address from non-validated to bidirectional links
            if sender_address in self.non_validated_neighbors:
                self.non_validated_neighbors.discard(sender_address)
            self.valid_bidirectional_links[sender_address] = valid_links
            # update information in the neighbor table as well
            self.neighbor_table[sender_address] = {
                'status': OLSRLinkStatus.Bidirectional,
                'two_hop': valid_links}
        # the alternative is that the bidirectional
        # verification is not (yet) completed. add the sender
        # to the non-validated neighbors list.
        else:
            # demote from bidirectional links if exists
            if sender_address in self.valid_bidirectional_links:
                self.valid_bidirectional_links.pop(sender_address)
            self.non_validated_neighbors.add(sender_address)
            # update information in the neighbor table as well
            self.neighbor_table[sender_address] = {
                'status': OLSRLinkStatus.Unidirectional,
                'two_hop': valid_links}

    def select_mpr(self):
        prev_mpr_set = self.mpr_set
        self.mpr_set = obtain_mpr(self.valid_bidirectional_links)
        if prev_mpr_set == self.mpr_set:
            self.mpr_sequence_number += 1

    def update_ms(self, sender_address, sender_mpr):
        # if we are in the sender's MPR set,
        # add the neighbor's address to the MS
        # set and increase the sequence number.
        if self.address in sender_mpr:
            if sender_address not in self.ms_set:
                self.ms_set.add(sender_address)
                self.ms_sequence_number += 1
        # if not; but we used to be there; then
        # the neighbor node discarded us as an MPR.
        # remove the neighbor from the MS set
        # and increase the sequence number
        elif sender_address in self.ms_set:
            self.ms_set.discard(sender_address)
            self.ms_sequence_number += 1

    def on_tc_handler(self, payload):
        self.on_tc(payload['address'], payload['origin'],
                   payload['ms_set'], payload['ms_seq_num'])

    # sender_address is the message originator.
    # last_hop is the node that passed us the tc message.
    def on_tc(self, last_hop, sender_address, sender_ms, sender_ms_sequence_number):
        # discard our own TC messages
        if sender_address == self.address:
            return

        # discard tc messages with an old sequence number
        if sender_address in self.tc_seq_cache:
            if self.tc_seq_cache[sender_address] >= sender_ms_sequence_number:
                return
        self.tc_seq_cache[sender_address] = sender_ms_sequence_number

        # we are looking for entries whose destination is the sender of the TC message
        for (_, destination), (sequence_number, _) in self.topology_table.items():
            # if the MS sequence number is already higher;
            # we can discard the message and return
            if destination == sender_address and sequence_number > sender_ms_sequence_number:
                return

        # otherwise, we remove each entry for the sender
        # address as destination, to add them later
        keys_to_pop = []
        for k in self.topology_table.keys():
            (destination, _) = k
            if destination == sender_address:
                keys_to_pop.append(k)
        for k in keys_to_pop:
            self.topology_table.pop(k)

        # insert the new topology data to the table
        current_time = datetime.now()
        for dst in sender_ms:
            self.topology_table[(dst, sender_address)] = (sender_ms_sequence_number, current_time)

        # update the routing table
        self.calculate_route_table()

        # if this node is an MPR of the last hop; forward the message
        if last_hop in self.ms_set:
            self.broadcast_msg(HOLSRMessageTypes.TC, OLSRMessagePayload({
                'address': self.address, 'origin': sender_address,
                'ms_set': sender_ms, 'ms_seq_num': sender_ms_sequence_number}))

    def calculate_route_table(self):
        # start populating a new routing table by
        # adding one-hop neighbors to hop distance 1
        routes = {}
        for one_hop in self.valid_bidirectional_links.keys():
            routes[one_hop] = (one_hop, 1)
        # we will iterate the topology table keys (which are node pairs)
        # to add routes at hop_distance away at each iteration.
        # once no changes occur; we break the loop.
        hop_distance = 0
        loop = True
        while loop:
            loop = False
            hop_distance += 1
            for (destination, last_hop) in self.topology_table.keys():
                if destination != self.address and destination not in routes:
                    if last_hop in routes:
                        (next_hop, distance) = routes.get(last_hop)
                        if distance == hop_distance:
                            routes[destination] = (next_hop, distance + 1)
                            loop = True
        self.routing_table = routes

    def get_route(self, destination):
        if destination not in self.routing_table:
            return None
        (next_hop, _) = self.routing_table.get(destination)
        return next_hop

    def send_data_to(self, sender, destination, data):
        # if the routing is not yet complete;
        # drop the message.
        next_hop = self.get_route(destination)
        if next_hop is not None:
            self.broadcast_msg(HOLSRMessageTypes.DATA, OLSRMessagePayload({
                'address': sender, 'destination': destination,
                'next_hop': next_hop, 'data': data}))

    def on_data_handler(self, payload):
        self.on_data_fwd(payload['address'], payload['destination'],
                         payload['next_hop'], payload['data'])

    def on_data_fwd(self, sender, destination, me, data):
        # drop messages not forwarded to us. normally,
        # if we did p2p we wouldn't need this step.
        # but to avoid our virtual address to node ids;
        # we send these messages to all our neighbors,
        # and then drop it here.
        if me == self.address:
            if destination == self.address:
                # if we're here; the message was directed to us.
                self.on_data(sender, data)
            else:
                # we were indeed the next hop; but not the
                # ultimate destination. keep forwarding the
                # mail using the routing table.
                self.send_data_to(sender, destination, data)

    def on_data(self, sender, data):
        # TODO
        pass


class HOLSRComponent(OLSRComponent):
    def __init__(self, component_name, component_id, context=None,
                 configurationparameters=None, num_worker_threads=1):
        super().__init__(component_name, component_id, context=context,
                         configurationparameters=configurationparameters,
                         num_worker_threads=num_worker_threads)
        layers_set = {0}
        if configurationparameters is not None:
            if configurationparameters['layers'] is not None:
                layers_set = configurationparameters['layers']
        is_head = len(layers_set) > 1
        self.layers = layers_set
        self.cluster_head = self.address if is_head else ''
        self.cluster_dist = 0
        self.last_cia = datetime.now()
        # members of the cluster (for cluster heads)
        self.cluster_member_table = {}
        # decorate old message handlers to discard messages from different clusters,
        # add HOLSR-specific message handlers,
        # then decorate both old and new handlers to discard messages from different layers
        self.message_handlers[HOLSRMessageTypes.CIA] = lambda p: self.on_cia_handler(p)
        self.message_handlers[HOLSRMessageTypes.HTC] = lambda p: self.on_htc_handler(p)

    def on_hello_handler(self, payload):
        if len(self.layers.intersection(payload['layers'])) > 0:
            if payload['cluster_head'] == self.cluster_head:
                super().on_hello_handler(payload)

    def on_tc_handler(self, payload):
        if len(self.layers.intersection(payload['layers'])) > 0:
            if payload['cluster_head'] == self.cluster_head:
                super().on_tc_handler(payload)

    def on_time(self, current_time, delta):
        super().on_time(current_time, delta)
        if (current_time - self.last_cia).total_seconds() > 0.25:
            self.last_cia = current_time
            self.broadcast_cia()

    def broadcast_msg(self, msg_type, payload):
        # decorate all outgoing messages with cluster and layer info
        if 'layers' not in payload.messagepayload:
            payload.messagepayload['layers'] = self.layers
        if 'cluster_head' not in payload.messagepayload:
            payload.messagepayload['cluster_head'] = self.cluster_head
        super().broadcast_msg(msg_type, payload)

    def broadcast_cia(self):
        # only cluster heads broadcast initial CIA messages. others relay
        if self.cluster_head == self.address:
            self.broadcast_msg(HOLSRMessageTypes.CIA, OLSRMessagePayload({
                'address': self.address, 'cluster_head': self.address, 'distance': 0}))

    def on_cia_handler(self, payload):
        self.on_cia(payload['address'], payload['layers'],
                    payload['cluster_head'], payload['distance'])

    def on_cia(self, sender_address, layers, sender_head, sender_distance):
        # discard our own CIA messages
        if self.address == sender_head:
            return
        # CIA msgs of lower levels or this level contribute to the cluster member table
        if self.cluster_head == self.address:
            if self.is_from_lower_levels(layers):
                inserted = False
                if sender_head not in self.cluster_member_table:
                    self.cluster_member_table[sender_head] = []
                else:
                    for i in range(len(self.cluster_member_table[sender_head])):
                        (member_addr, member_hop) = self.cluster_member_table[sender_head][i]
                        if member_addr == sender_address:
                            self.cluster_member_table[sender_head][i] = (sender_address, sender_distance)
                            inserted = True
                            break
                if not inserted:
                    self.cluster_member_table[sender_head].append((sender_address, sender_distance))
        # for cluster selection; don't consider messages from other layers
        if len(self.layers.intersection(layers)) == 0:
            return
        # our distance is one hop more than the sender's
        distance = sender_distance + 1
        # if the first CIA msg, or a closer cluster head is seen
        if self.cluster_head == '' or distance < self.cluster_dist:
            # update the cluster head and distance
            self.cluster_head = sender_head
            self.cluster_dist = distance
            # propagate the message
            self.broadcast_msg(HOLSRMessageTypes.CIA, OLSRMessagePayload({
                'address': self.address, 'cluster_head': sender_head, 'distance': distance}))

    def is_from_lower_levels(self, sender_layers):
        for self_layer in self.layers:
            for sender_layer in sender_layers:
                if sender_layer <= self_layer:
                    return True
        return False

    def on_htc_handler(self, payload):
        self.on_htc(payload['layers'], payload['address'], payload['members'])

    def on_htc(self, _layers, cluster_head, cluster_member_list):
        # Only cluster heads maintain a member table
        if self.cluster_head == self.address:
            self.cluster_member_table[cluster_head] = cluster_member_list

    def broadcast_htc(self):
        # only cluster heads broadcast HTC messages
        if self.cluster_head == self.address:
            payload = {'address': self.address, 'members': []}
            self.lock.acquire()
            try:
                if self.address in self.cluster_member_table:
                    payload['members'] = [a for a in self.cluster_member_table[self.address]]
            finally:
                self.lock.release()
            self.broadcast_msg(HOLSRMessageTypes.HTC, OLSRMessagePayload(payload))





import os
import time
import pickle
import random
import pandas as pd
import seaborn as sns
import networkx as nx
import matplotlib.pyplot as plt
from datetime import datetime as dt

from ahc.Ahc import Topology
from graph import ERG, Grid, Star
from ahc.TerminationDetection.dijkstra_scholten import DijkstraScholtenAdHocNode, DSAHCNodeSimulationStatus
from ahc.Channels.Channels import P2PFIFOPerfectChannel

import glob
from PIL import Image

# # filepaths
# fp_in = "/path/to/image_*.png"
# fp_out = "/path/to/image.gif"

def make_gif(from_path: str, to_path: str, out_file: str, duration: int) -> None:
    # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
    _images = sorted(glob.glob(from_path))
    print(_images)
    img, *imgs = [Image.open(f) for f in _images]
    img.save(fp=f"{to_path}/{out_file}", format='GIF', append_images=imgs,
            save_all=True, duration=duration, loop=0)

def run_dijkstra_scholten_simulation(args):
    ts = dt.now().timestamp()

    if args.network_type == "erg":
        N = ERG(args.node_count, args.node_connectivity)
        total_nodes = args.node_count
    elif args.network_type == "grid":
        N = Grid(args.node_count_on_edge)
        total_nodes = args.node_count_on_edge ** 2
    elif args.network_type == "star":
        N = Star(args.slave_count, master_is_root=args.master_is_root)
        total_nodes = args.slave_count + 1

    if args.save_tick_plots:
        tick_plots_save_path = f"{args.tick_plots_save_dir}/DS_{args.network_type}_{total_nodes}_{ts}/"
        os.mkdir(tick_plots_save_path)

        print(f"++ Tick plots will be saved to: {tick_plots_save_path}")

    if args.run_until_termination:
        args.simulation_ticks = 10**10

    assert args.hard_stop_max_tick < args.simulation_ticks
    assert args.hard_stop_min_tick > 0

    hard_stop_on_tick = None

    if args.hard_stop_nodes:
        hard_stop_on_tick = [random.randint(args.hard_stop_min_tick, args.hard_stop_max_tick) for _ in range(total_nodes)]

    node_active_ticks_initial = [random.randint(args.node_min_activeness_after_receive, args.node_max_activeness_after_receive) if random.random() <= args.node_initial_activeness_prob else 0 for _ in range(total_nodes)]
    alive_nodes = list(range(total_nodes))
    
    topo_context = {
        "network": N,
        "ms_per_tick": args.ms_per_tick,
        "simulation_ticks": args.simulation_ticks,
        "initial_liveness": node_active_ticks_initial,
        "communication_on_active_prob": args.node_activeness_communication_prob,
        "min_activeness_after_receive": args.node_min_activeness_after_receive,
        "max_activeness_after_receive": args.node_max_activeness_after_receive,
        "node_package_process_per_tick": args.node_package_process_per_tick,
        "passiveness_death_thresh": args.passiveness_death_thresh,
        "hard_stop_on_tick": hard_stop_on_tick,
        "alive_nodes": alive_nodes,
        "only_root_alive_initially": args.only_root_alive_initially
    }

    print(topo_context)

    topo = Topology()
    topo.construct_from_graph(N.G, DijkstraScholtenAdHocNode, P2PFIFOPerfectChannel, context=topo_context)

    topo.start()

    stats = {
        "df": pd.DataFrame(data={
            "dead_nodes": [],
            "active_nodes": [],
            "packets_in_transmit": [],
            "queued_packets": [],
            "control_packets_sent": [],
            "control_basic_instant_ratio": [],
            "control_total_cumulative_ratio": [],
        }),
        "terminated_on_tick": None,
        "announced_on_tick": None
    }

    graphs = []

    fig, axes = plt.subplots(2, 3)
    fig.set_figwidth(25)
    fig.set_figheight(10)
    # fig.tight_layout()

    input("\n>>> Proceed ?")

    term_wait_ctr = 0
    reason = None

    try:
        for t in range(1, args.simulation_ticks + 1):
            print(f"[S] Tick: {t}")

            packages_sent = 0
            packages_waiting_on_queue = 0
            num_nodes_active = 0
            num_dead_nodes = 0
            break_this_tick = False
            control_packets_sent = 0

            T = nx.Graph()
            node_color = []

            for index, node in topo.nodes.items():
                new_state, pkg_sent_to_friend, cps = node.simulation_tick()

                if N.root == index:
                    node_color.append("red")
                elif new_state == DSAHCNodeSimulationStatus.ACTIVE:
                    node_color.append("green")
                elif new_state == DSAHCNodeSimulationStatus.PASSIVE:
                    node_color.append("mediumslateblue")
                elif new_state == DSAHCNodeSimulationStatus.OUT_OF_TREE:
                    node_color.append("gray")

                if index not in T.nodes():
                    T.add_node(index)

                if node.parent_node is not None:
                    if node.parent_node not in T.nodes():
                        T.add_node(node.parent_node)

                    T.add_edge(index, node.parent_node)

                if index == N.root and new_state is None:
                    reason = f"root terminated ({t})"
                    break_this_tick = True

                if new_state == DSAHCNodeSimulationStatus.ACTIVE:
                    num_nodes_active += 1
                elif new_state == DSAHCNodeSimulationStatus.OUT_OF_TREE:
                    num_dead_nodes += 1

                if pkg_sent_to_friend is not None:
                    packages_sent += 1

                control_packets_sent += cps
                packages_waiting_on_queue += node.waiting_packages_on_queue

            print(f"   (ACTIVE: {num_nodes_active}, PKGS-WAIT: {packages_waiting_on_queue}, PKGS-SENT: {packages_sent})")

            if (packages_waiting_on_queue == 0 and num_nodes_active == 0 and packages_sent == 0):
                if stats["terminated_on_tick"] is None: 
                    stats["terminated_on_tick"] = t

                print("!!! TERMINATED !!!")

                term_wait_ctr += 1

                if args.wait_ticks_after_termination > 0 and term_wait_ctr > args.wait_ticks_after_termination:
                    print("!!! FORCE TERMINATED !!!")
                    reason = f"forced ({args.wait_ticks_after_termination})"
                    break_this_tick = True

                if args.exit_on_termination:
                    break_this_tick = True

            total_pkgs_sent_cum = (stats["df"]["control_packets_sent"].sum() + control_packets_sent + stats["df"]["packets_in_transmit"].sum() + packages_sent)

            stats["df"].loc[t-1] = [
                num_dead_nodes,
                num_nodes_active,
                packages_sent,
                packages_waiting_on_queue,
                control_packets_sent,
                (control_packets_sent / packages_sent) if packages_sent > 0 else 0, # TODO: Fix later, find a better soln,,,
                (((stats["df"]["control_packets_sent"].sum() + control_packets_sent) / total_pkgs_sent_cum) if total_pkgs_sent_cum > 0 else 0) * 100
            ]

            graphs.append({
                "edges": T.edges(),
                "nodes": T.nodes(),
            })
            
            if not args.no_realtime_plot or args.save_tick_plots:
                axes[0][0].cla()
                axes[0][1].cla()
                axes[0][2].cla()
                axes[1][0].cla()
                axes[1][1].cla()
                axes[1][2].cla()

                axes[0][0].set_title("Active/Dead Nodes")
                axes[0][0].set_xlabel(f"Simulation Ticks ({args.ms_per_tick}ms/tick)")
                axes[0][0].set_ylabel(f"Node Count")

                axes[0][1].set_title("Packets in Transmit")
                axes[0][1].set_xlabel(f"Simulation Ticks ({args.ms_per_tick}ms/tick)")
                axes[0][1].set_ylabel(f"Packet Count")

                axes[1][0].set_title("Control Packet Ratio")
                axes[1][0].set_xlabel(f"CPR ({args.ms_per_tick}ms/tick)")
                axes[1][0].set_ylabel(f"CPR (control packets/total packets)")

                axes[1][1].set_title("Instant Control Packet Ratio")
                axes[1][1].set_xlabel(f"Simulation Ticks ({args.ms_per_tick}ms/tick)")
                axes[1][1].set_ylabel(f"CPR (control packets/total packets)")

                sns.lineplot(data=stats["df"]["active_nodes"], ax=axes[0][0], color="orange", legend='brief', label="Active nodes")
                sns.lineplot(data=stats["df"]["dead_nodes"], ax=axes[0][0], color="blue", legend='brief', label="Dead nodes")
                sns.lineplot(data=stats["df"]["packets_in_transmit"], ax=axes[0][1], color="purple", legend='brief', label="Basic packets")
                sns.lineplot(data=stats["df"]["control_packets_sent"], ax=axes[0][1], color="green", legend='brief', label="Control packets")
                sns.lineplot(data=stats["df"]["control_total_cumulative_ratio"], ax=axes[1][0], color="mediumslateblue")
                sns.lineplot(data=stats["df"]["control_basic_instant_ratio"], ax=axes[1][1], color="red")

                nx.draw(T, ax=axes[0][2], with_labels=True, node_color=node_color)
                
                if args.network_type == "grid":   
                    pos = {i: (i // args.node_count_on_edge, i % args.node_count_on_edge) for i in range(total_nodes)}
                    nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color, pos=pos)
                else:
                    nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color)

                plt.pause(0.0005)

                if args.save_tick_plots:
                    plt.savefig(tick_plots_save_path + f"{str(t).zfill(3)}.png", dpi=160)

            if break_this_tick:
                break

            time.sleep(args.ms_per_tick / 1000)
    except KeyboardInterrupt:
        pass

    axes[0][0].cla()
    axes[0][1].cla()
    axes[0][2].cla()
    axes[1][0].cla()
    axes[1][1].cla()
    axes[1][2].cla()

    sns.lineplot(data=stats["df"]["active_nodes"], ax=axes[0][0], color="orange")
    sns.lineplot(data=stats["df"]["dead_nodes"], ax=axes[0][0], color="blue")
    sns.lineplot(data=stats["df"]["packets_in_transmit"], ax=axes[0][1], color="purple")
    sns.lineplot(data=stats["df"]["control_packets_sent"], ax=axes[0][1], color="green")
    sns.lineplot(data=stats["df"]["control_total_cumulative_ratio"], ax=axes[1][0], color="mediumslateblue")
    sns.lineplot(data=stats["df"]["control_basic_instant_ratio"], ax=axes[1][1], color="red")
    
    nx.draw(T, ax=axes[0][2], with_labels=True, node_color=node_color)
    
    if args.network_type == "grid":   
        pos = {i: (i // args.node_count_on_edge, i % args.node_count_on_edge) for i in range(total_nodes)}
        nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color, pos=pos)
    else:
        nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color)

    plt.savefig(f"simdump/DS_stats_{args.network_type}_{total_nodes}_{ts}.png", dpi=200)

    with open(f"simdump/DS_run_{args.network_type}_{total_nodes}_{ts}.pkl", "wb") as fp:
        pickle.dump({
            "args": args,
            "context": topo_context,
            "stats": stats,
            "graphs": graphs
        }, fp)

    if not args.no_realtime_plot:
        plt.show()

    print(f"\n{reason} [{t - stats['terminated_on_tick'] if stats['terminated_on_tick'] is not None else None}]")

    if args.generate_gif:
        make_gif(f"{tick_plots_save_path}/*.png", tick_plots_save_path, "animation.gif", t * 20)


from ahc.Routing.ARA.ARAConfiguration import ARAConfiguration

class ARARoutingEntry:
  def __init__(self, destination, nextHopAddress):
    self.destination = destination
    self.nextHopAddress = nextHopAddress
    self.pheromone = 0.0

  def increasePheromone(self):
    self.pheromone = self.pheromone + ARAConfiguration.PHEROMONE_DELTA

  def evaporatePheromone(self):
    self.pheromone = (1-ARAConfiguration.EVAPORATION_FACTOR) * self.pheromone



import queue
import random
from enum import Enum
from threading import Thread

from ahc.Ahc import ComponentModel, EventTypes, ConnectorList, MessageDestinationIdentifiers
from ahc.Ahc import Event

# TODO: Channel failure models: lossy-link, fair-loss, stubborn links, perfect links (WHAT ELSE?), FIFO perfect
# TODO: Logged perfect links (tolerance to crashes), authenticated perfect links
# TODO: Broadcast channels? and their failure models? Collisions?
# TODO: Properties of all models separately: e.g., Fair loss, finite duplication, no fabrication in fair-loss link model
# TODO: Packets: loss, duplication, sequence change, windowing?,
# TODO: Eventually (unbounded time) or bounded time for message delivery?


# Channels have three events: sendtochannel, processinchannel and delivertocomponent
# Components tell channels to handle a message by the EventTypes.MFRT event, the component calls senddown with the event EventTypes.MFRT
# First pipeline stage moves the message to the interim pipeline stage with the "processinchannel" event for further processing, such as  the channel may drop it, delay it, or whatever
# Channels deliver the message to output queue by the "delivertocomponent" event
# The output queue then will send the message up to the connected component(s) using the "messagefromchannel" event
# The components that will use the channel directly, will have to handle "messagefromchannel" event

class ChannelEventTypes(Enum):
  INCH = "processinchannel"
  DLVR = "delivertocomponent"

class Channel(ComponentModel):

  def on_init(self, eventobj: Event):

    pass

  # Overwrite onSendToChannel if you want to do something in the first pipeline stage
  def on_message_from_top(self, eventobj: Event):
    # channel receives the input message and will process the message by the process event in the next pipeline stage
    # Preserve the event id through the pipeline
    myevent = Event(eventobj.eventsource, ChannelEventTypes.INCH,
                    eventobj.eventcontent, eventid=eventobj.eventid)
    self.channelqueue.put_nowait(myevent)

  # Overwrite onProcessInChannel if you want to do something in interim pipeline stage
  def on_process_in_channel(self, eventobj: Event):
    # Add delay, drop, change order whatever....
    # Finally put the message in outputqueue with event deliver
    # Preserve the event id through the pipeline
    myevent = Event(eventobj.eventsource, ChannelEventTypes.DLVR,
                    eventobj.eventcontent, eventid=eventobj.eventid)
    self.outputqueue.put_nowait(myevent)

  # Overwrite onDeliverToComponent if you want to do something in the last pipeline stage
  # onDeliver will deliver the message from the channel to the receiver component using messagefromchannel event
  def on_deliver_to_component(self, eventobj: Event):
    callername = eventobj.eventsource.componentinstancenumber
    for item in self.connectors:
      callees = self.connectors[item]
      for callee in callees:
        calleename = callee.componentinstancenumber
        # print(f"I am connected to {calleename}. Will check if I have to distribute it to {item}")
        if calleename == callername:
          pass
        else:
          # Preserve the event id through the pipeline
          myevent = Event(eventobj.eventsource, EventTypes.MFRB,
                          eventobj.eventcontent, self.componentinstancenumber,
                          eventid=eventobj.eventid)
          callee.trigger_event(myevent)

  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.outputqueue = queue.Queue()
    self.channelqueue = queue.Queue()
    self.eventhandlers[ChannelEventTypes.INCH] = self.on_process_in_channel
    self.eventhandlers[ChannelEventTypes.DLVR] = self.on_deliver_to_component

    for i in range(self.num_worker_threads):
      # note that the input queue is handled by the super class...
      t = Thread(target=self.queue_handler, args=[self.channelqueue])
      t1 = Thread(target=self.queue_handler, args=[self.outputqueue])
      t.daemon = True
      t1.daemon = True
      t.start()
      t1.start()


class BasicLossyChannel(Channel):
  def __init__(self, componentname, componentinstancenum, loss_percentage=0):
    super().__init__(componentname, componentinstancenum,)
    self.loss_percentage = loss_percentage
  # Overwrite onDeliverToComponent if you want to do something in the last pipeline stage
  # onDeliver will deliver the message from the channel to the receiver component using messagefromchannel event
  def on_deliver_to_component(self, eventobj: Event):
    callername = eventobj.eventsource.componentinstancenumber
    for item in self.connectors:
      callees = self.connectors[item]
      for callee in callees:
        calleename = callee.componentinstancenumber
        # print(f"I am connected to {calleename}. Will check if I have to distribute it to {item}")
        if calleename == callername:
          pass
        else:
          randomnum = random.uniform(0, 1)
          # print("random number here is " , randomnum, " for ", self.componentinstancenumber)
          if randomnum >= self.loss_percentage:
            myevent = Event(eventobj.eventsource, EventTypes.MFRB, eventobj.eventcontent, self.componentinstancenumber)
            callee.trigger_event(myevent)


class AHCChannelError(Exception):
  pass

class P2PFIFOPerfectChannel(Channel):

  # Overwrite onSendToChannel
  # Channels are broadcast, that is why we have to check channel id's using hdr.interfaceid for P2P
  def on_message_from_top(self, eventobj: Event):
    # if channelid != hdr.interfaceif then drop (should not be on this channel)
    hdr = eventobj.eventcontent.header
    if hdr.nexthop != MessageDestinationIdentifiers.LINKLAYERBROADCAST:
      if set(hdr.interfaceid.split("-")) == set(self.componentinstancenumber.split("-")):
        #print(f"Will forward message since {hdr.interfaceid} and {self.componentinstancenumber}")
        myevent = Event(eventobj.eventsource, ChannelEventTypes.INCH, eventobj.eventcontent)
        self.channelqueue.put_nowait(myevent)
      else:
        #print(f"Will drop message since {hdr.interfaceid} and {self.componentinstancenumber}")
        pass

  def on_deliver_to_component(self, eventobj: Event):
    msg = eventobj.eventcontent
    callername = eventobj.eventsource.componentinstancenumber
    for item in self.connectors:
      callees = self.connectors[item]
      for callee in callees:
        calleename = callee.componentinstancenumber
        # print(f"I am connected to {calleename}. Will check if I have to distribute it to {item}")
        if calleename == callername:
          pass
        else:
          myevent = Event(eventobj.eventsource, EventTypes.MFRB, eventobj.eventcontent, self.componentinstancenumber)
          callee.trigger_event(myevent)

  # Overwriting to limit the number of connected components
  def connect_me_to_component(self, name, component):
    try:
      self.connectors[name] = component
      # print(f"Number of nodes connected: {len(self.ports)}")
      if len(self.connectors) > 2:
        raise AHCChannelError("More than two nodes cannot connect to a P2PFIFOChannel")
    except AttributeError:
      self.connectors = ConnectorList()
      self.connectors[name] = component
    # except AHCChannelError as e:
    #    print( f"{e}" )

class P2PFIFOFairLossChannel(P2PFIFOPerfectChannel):
  prob = 1
  duplicationprobability = 0
  # Overwrite onSendToChannel
  # Channels are broadcast, that is why we have to check channel id's using hdr.interfaceid for P2P

  def on_message_from_top(self, eventobj: Event):
    # if channelid != hdr.interfaceif then drop (should not be on this channel)
    hdr = eventobj.eventcontent.header
    if hdr.nexthop != MessageDestinationIdentifiers.LINKLAYERBROADCAST:
      if set(hdr.interfaceid.split("-")) == set(self.componentinstancenumber.split("-")):
        #print(f"Will forward message since {hdr.interfaceid} and {self.componentinstancenumber}")
        myevent = Event(eventobj.eventsource, ChannelEventTypes.INCH, eventobj.eventcontent)
        self.channelqueue.put_nowait(myevent)
      else:
        #print(f"Will drop message since {hdr.interfaceid} and {self.componentinstancenumber}")
        pass


  def on_process_in_channel(self, eventobj: Event):
    if random.random() < self.prob:
      myevent = Event(eventobj.eventsource, ChannelEventTypes.DLVR, eventobj.eventcontent)
      self.outputqueue.put_nowait(myevent)
    if random.random() < self.duplicationprobability:
      self.channelqueue.put_nowait(eventobj)

  def setPacketLossProbability(self, prob):
    self.prob = prob

  def setAverageNumberOfDuplicates(self, d):
    if d > 0:
      self.duplicationprobability = (d - 1) / d
    else:
      self.duplicationprobability = 0

class FIFOBroadcastPerfectChannel(Channel):
  pass



import time

from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes, ComponentRegistry, Lock, \
    Topology
from enum import Enum
import datetime
import threading
from ahc.Routing.CGSR.CGSR import RouteRequestPacket, RREPPacket, SourceRouting, HelloMessage, ClusterAdjacencyExtension, CGSREventTypes, \
    CGSRMessageType
import random


class CGSRControlThread(threading.Thread):
    def __init__(self, event, interval, unique_name, component):
        threading.Thread.__init__(self)
        self.stopped = event
        self.increment = 0
        self.increment2 = 0
        self.interval = interval
        self.unique_name = unique_name
        self.component = component
        self.wakeup=True

    def run(self):
        if self.wakeup is True:
            time.sleep(2)
        while not self.stopped.wait(1):
            self.wakeup=False
            # print(f"control thread for node {self.unique_name} has started")
            self.increment += 1
            if self.increment == 2:
                self.component.generate_hello()

            if self.increment == 4:
                self.component.generate_hello()
                self.component.generate_rreq()
                self.increment = 0


class CGSRControlComponent(ComponentModel):

    def __init__(self, componentname, componentid):
        super(CGSRControlComponent, self).__init__(componentname, componentid)
        self.C_TIMER = 4
        self.U_TIMER = 3
        self.HELLO_INTERVAL = 2
        self.MAX_JITTER = self.HELLO_INTERVAL / 4
        self.JITTER = round(random.uniform(0, self.MAX_JITTER))
        self.RANDOM_HELLO_INTERVAL = self.HELLO_INTERVAL - self.MAX_JITTER

        self.timer = threading.Event()
        self.control_thread = CGSRControlThread(self.timer, self.RANDOM_HELLO_INTERVAL, self.unique_name(), self)
        self.send_self(Event(self, EventTypes.INIT, None))

    def on_init(self, eventobj: Event):
        super(CGSRControlComponent, self).on_init(eventobj)
        #print('control thread started')

        self.control_thread.start()

    def generate_hello(self):
        self.send_peer(Event(self, EventTypes.MFRP, CGSRMessageType.GENERATE_HELLO))
        # print('generate hello')

    def generate_rreq(self):
        self.send_peer(Event(self, EventTypes.MFRP, CGSRMessageType.GENERATE_RREQ))



from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes, Thread, Lock


# where the machine learning model is loaded... The top entity for the Node...
class ApplicationComponent(ComponentModel):
    def __init__(self, componentname, componentid):
        super(ApplicationComponent, self).__init__(componentname, componentid)

        pass

    def job(self, *args):

        while True:
            data = input("Information to send : ")
            message_header = GenericMessageHeader("APPQUERY",
                                                  "ApplicationComponent-" + str(self.componentinstancenumber),
                                                  "Coordinator-" + str(self.componentinstancenumber))
            message = GenericMessage(message_header, (5, data))
            kickstarter = Event(self, EventTypes.MFRT, message)
            self.send_down(kickstarter)


    def on_init(self, eventobj: Event):
        if self.componentinstancenumber == 0:
            message_header = GenericMessageHeader("INITIATE", "ApplicationComponent-"+str(self.componentinstancenumber),
                                                  "Coordinator-" + str(self.componentinstancenumber))
            message = GenericMessage(message_header, "")
            kickstarter = Event(self, EventTypes.MFRT, message)
            self.send_down(kickstarter)
            print(f"App {self.componentinstancenumber} sends an INITIATE to Coordinator")

            thread = Thread(target=self.job, args=[45, 54, 123])
            thread.start()

    def on_message_from_bottom(self, eventobj: Event):
        sender = eventobj.eventcontent.header.messagefrom.split("-")[0]
        messageto = eventobj.eventcontent.header.messageto.split("-")[0]
        message_type = eventobj.eventcontent.header.messagetype
        message = eventobj.eventcontent.payload
        if messageto == "ApplicationComponent":
            if message_type == "APPQUERY":
                source, content = message
                print(f"App {self.componentinstancenumber} has received {message} from {source}")
                message_header = GenericMessageHeader("APPRESPONSE",
                                                      "ApplicationComponent-" + str(self.componentinstancenumber),
                                                      "Coordinator-" + str(self.componentinstancenumber))
                message = GenericMessage(message_header, (source, "Hellooooooooo "+content))
                kickstarter = Event(self, EventTypes.MFRT, message)
                self.send_down(kickstarter)

            elif message_type == "APPRESPONSE":
                source, content = message
                print(f"App {self.componentinstancenumber} has received APPRESPONSE {message} from {source}")



#!/usr/bin/env python
""" Implements the AHC library.

TODO: Longer description of this module to be written.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.

"""

__author__ = "One solo developer"
__authors__ = ["Ertan Onur", "Berke Tezergil", "etc"]
__contact__ = "eonur@ceng.metu.edu.tr"
__copyright__ = "Copyright 2021, WINSLAB"
__credits__ = ["Ertan Onur", "Berke Tezergil", "etc"]
__date__ = "2021/04/07"
__deprecated__ = False
__email__ = "eonur@ceng.metu.edu.tr"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"

import datetime
import queue
from enum import Enum
from threading import Thread, Lock
from random import sample, random, sample
import itertools
from matplotlib.pyplot import flag
import networkx as nx
import itertools
import threading
import time
# TIMING ASSUMPTIONS
# TODO: Event handling time, message sending time, assumptions about clock (drift, skew, ...)
# TODO: 1. Asynch,  2. Synch 3. Partial-synch 4. Timed asynch
# TODO: Causal-order (happen before), total-order,
# TODO: Causal-order algebra!!!
# TODO: Implement logical clocks (Lamport clocks, vector clocks) in event handling loop

#  AUTOMATA and EXECUTIONS
# TODO: Let component model hande executions and chekcs on executions (which event, in which order, per process or per system, similarity of executions)

#  VISUALIZATION
# TODO: Space-time diagrams for events

#  TOPOLOGY MANAGEMENT
# TODO: Given a graph as input, generate the topology....

inf = float('inf')


# The following are the common default events for all components.
class EventTypes(Enum):
  INIT = "init"
  MFRB = "msgfrombottom"
  MFRT = "msgfromtop"
  MFRP = "msgfrompeer"


class MessageDestinationIdentifiers(Enum):
  LINKLAYERBROADCAST = -1,  # sinngle-hop broadcast, means all directly connected nodes
  NETWORKLAYERBROADCAST = -2  # For flooding over multiple-hops means all connected nodes to me over one or more links


# A Dictionary that holds a list for the same key
class ConnectorList(dict):

  def __setitem__(self, key, value):
    try:
      self[key]
    except KeyError:
      super(ConnectorList, self).__setitem__(key, [])
    self[key].append(value)


class ConnectorTypes(Enum):
  DOWN = "DOWN"
  UP = "UP"
  PEER = "PEER"


def auto_str(cls):

    def __str__(self):
        return '%s(%s)' % (
            type(self).__name__,
            ', '.join('%s=%s' % item for item in vars(self).items())
        )

    cls.__str__ = __str__
    return cls


@auto_str
class GenericMessagePayload:

  def __init__(self, messagepayload):
    self.messagepayload = messagepayload


@auto_str
class GenericMessageHeader:

  def __init__(self, messagetype, messagefrom, messageto, nexthop=float('inf'), interfaceid=float('inf'), sequencenumber=-1):
    self.messagetype = messagetype
    self.messagefrom = messagefrom
    self.messageto = messageto
    self.nexthop = nexthop
    self.interfaceid = interfaceid
    self.sequencenumber = sequencenumber


@auto_str
class GenericMessage:

  def __init__(self, header, payload):
    self.header = header
    self.payload = payload
    self.uniqueid = str(header.messagefrom) + "-" + str(header.sequencenumber)


@auto_str
class Event:
  curr_event_id = 0

  def __init__(self, eventsource, event, eventcontent, fromchannel=None,
               eventid=-1):
    self.eventsource = eventsource
    self.event = event
    self.time = datetime.datetime.now()
    self.eventcontent = eventcontent
    self.fromchannel = fromchannel
    self.eventid = eventid
    if self.eventid == -1:
      self.eventid = self.curr_event_id
      self.curr_event_id += 1

  def __eq__(self, other) -> bool:
    if type(other) is not Event:
      return False

    return self.eventid == other.eventid

  def __hash__(self) -> int:
    return self.eventid


def singleton(cls):
  instance = [None]

  def wrapper(*args, **kwargs):
    if instance[0] is None:
      instance[0] = cls(*args, **kwargs)
    return instance[0]

  return wrapper


@singleton
class ComponentRegistry:
  components = {}

  def get_component_by_instance(self, instance):
    list_of_keys = list()
    list_of_items = self.components.items()
    for item in list_of_items:
      if item[1] == instance:
        list_of_keys.append(item[0])
    return list_of_keys

  def add_component(self, component):
    key = component.componentname + str(component.componentinstancenumber)
    self.components[key] = component

  def get_component_by_key(self, componentname, componentinstancenumber):
    key = componentname + str(componentinstancenumber)
    return self.components[key]

  def init(self):
    for itemkey in self.components:
      cmp = self.components[itemkey]
      cmp.inputqueue.put_nowait(Event(self, EventTypes.INIT, None))
      print("Initializing, ", cmp.componentname, ":", cmp.componentinstancenumber)

  def print_components(self):
    for itemkey in self.components:
      cmp = self.components[itemkey]
      print(f"I am {cmp.componentname}.{cmp.componentinstancenumber}")
      for i in cmp.connectors:
        connectedcmp = cmp.connectors[i]
        for p in connectedcmp:
          print(f"\t{i} {p.componentname}.{p.componentinstancenumber}")

  def get_non_channel_components(self):
    res = []
    for itemkey in self.components:
      cmp = self.components[itemkey]
      if cmp.componentname.find("Channel") != -1:
        continue
      res.append(cmp)
    return res


registry = ComponentRegistry()

class ComponentConfigurationParameters():
    pass

class ComponentModel:
  terminated = False

  def on_init(self, eventobj: Event):
    # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
    pass

  def on_message_from_bottom(self, eventobj: Event):
    print(f"{EventTypes.MFRB} {self.componentname}.{self.componentinstancenumber}")

  def on_message_from_top(self, eventobj: Event):
    print(f"{EventTypes.MFRT}  {self.componentname}.{self.componentinstancenumber}")

  def on_message_from_peer(self, eventobj: Event):
    print(f"{EventTypes.MFRP}  {self.componentname}.{self.componentinstancenumber}")

  def __init__(self, componentname, componentinstancenumber, context=None, configurationparameters=None, num_worker_threads=1):
    self.context = context
    self.configurationparameters = configurationparameters
    self.eventhandlers = {EventTypes.INIT: self.on_init, EventTypes.MFRB: self.on_message_from_bottom,
                          EventTypes.MFRT: self.on_message_from_top, EventTypes.MFRP: self.on_message_from_peer}
    # Add default handlers to all instantiated components.
    # If a component overwrites the __init__ method it has to call the super().__init__ method
    self.inputqueue = queue.Queue()
    self.componentname = componentname
    self.componentinstancenumber = componentinstancenumber
    self.num_worker_threads = num_worker_threads
    try:
      if self.connectors:
        pass
    except AttributeError:
      self.connectors = ConnectorList()

    self.registry = ComponentRegistry()
    self.registry.add_component(self)

    for i in range(self.num_worker_threads):
      t = Thread(target=self.queue_handler, args=[self.inputqueue])
      t.daemon = True
      t.start()

  def connect_me_to_component(self, name, component):
    try:
      self.connectors[name] = component
    except AttributeError:
      self.connectors = ConnectorList()
      self.connectors[name] = component

  def remove_connection_to_component_from_me(self, name, component):
    try:
      self.connectors[name].remove(component)
    except AttributeError:
      pass

  def remove_connection_to_channel_from_me(self, name, channel):
    try:
      self.connectors[name].remove(channel)
    except AttributeError:
      pass
    connectornameforchannel = self.componentname + str(self.componentinstancenumber)
    channel.remove_connection_to_component_from_me(connectornameforchannel, self)
    self.on_break_connection_to_channel(name, channel)

  def on_break_connection_to_channel(self, name, channel):
    print(f"Connection broken to channel: {name}:{channel.componentinstancenumber}")

  def connect_me_to_channel(self, name, channel):
    try:
      self.connectors[name] = channel
    except AttributeError:
      self.connectors = ConnectorList()
      self.connectors[name] = channel
    connectornameforchannel = self.componentname + str(self.componentinstancenumber)
    channel.connect_me_to_component(connectornameforchannel, self)
    self.on_connected_to_channel(name, channel)

  def on_connected_to_channel(self, name, channel):
    print(f"Connected to channel: {name}:{channel.componentinstancenumber}")

  def on_pre_event(self, event):
    pass

  def unique_name(self):
    return f"{self.componentname}.{self.componentinstancenumber}"

  def terminate(self):
    self.terminated = True

  def send_down(self, event: Event):
    try:
      for p in self.connectors[ConnectorTypes.DOWN]:
        p.trigger_event(event)
    except:
      pass

  def send_up(self, event: Event):
    try:
      for p in self.connectors[ConnectorTypes.UP]:
        p.trigger_event(event)
    except:
      pass

  def send_peer(self, event: Event):
    try:
      for p in self.connectors[ConnectorTypes.PEER]:
        p.trigger_event(event)
    except:
      pass

  def send_self(self, event: Event):
    self.trigger_event(event)

  # noinspection PyArgumentList
  def queue_handler(self, myqueue):
    while not self.terminated:
      workitem = myqueue.get()
      if workitem.event in self.eventhandlers:
        self.on_pre_event(workitem)
        self.eventhandlers[workitem.event](eventobj=workitem)  # call the handler
      else:
        print(f"Event Handler: {workitem.event} is not implemented")
      myqueue.task_done()

  def trigger_event(self, eventobj: Event):
    self.inputqueue.put_nowait(eventobj)


@singleton
class Topology:
  nodes = {}
  channels = {}


  def construct_winslab_topology_with_channels(self, nodecount, nodetype, channeltype, context=None):
    
    self.construct_winslab_topology_without_channels(nodecount, nodetype, context)
    
    pairs = list(itertools.permutations(range(nodecount), 2))
    print("Pairs", pairs)
    self.G.add_edges_from(pairs)
    edges = list(self.G.edges)
    for k in edges:
      ch = channeltype(channeltype.__name__, str(k[0]) + "-" + str(k[1]))
      self.channels[k] = ch
      self.nodes[k[0]].connect_me_to_channel(ConnectorTypes.DOWN, ch)
      self.nodes[k[1]].connect_me_to_channel(ConnectorTypes.DOWN, ch)


  def construct_winslab_topology_without_channels(self, nodecount, nodetype, context=None):
    
    self.G = nx.Graph()
    self.G.add_nodes_from(range(nodecount))  # TODO : Change depending on the 

    nodes = list(self.G.nodes)
    for i in nodes:
      cc = nodetype(nodetype.__name__, i)
      self.nodes[i] = cc


  def construct_winslab_topology_without_channels_for_docker(self, nodetype, id, context=None):
    
    self.G = nx.Graph()
    self.G.add_nodes_from(range(1))  # TODO : Change depending on the 

    nodes = list(self.G.nodes)
    cc = nodetype(nodetype.__name__, id)
    self.nodes[0] = cc

  #def construct_from_graph(self, G: nx.Graph, nodetype, channeltype, context=None):
  def construct_from_graph(self, G: nx.Graph, nodetype, channeltype, context=None, dynamic = False, path = "topology.txt"):
    self.G = G
    nodes = list(G.nodes('conf', None))
    edges = list(G.edges)
    for (i, cf) in nodes:
      if cf is None:
        cc = nodetype(nodetype.__name__, i)
      else:
        cc = nodetype(nodetype.__name__, i, configurationparameters=cf)
      self.nodes[i] = cc
    for k in edges:
      ch = channeltype(channeltype.__name__, str(k[0]) + "-" + str(k[1]))
      self.channels[k] = ch
      self.nodes[k[0]].connect_me_to_channel(ConnectorTypes.DOWN, ch)
      self.nodes[k[1]].connect_me_to_channel(ConnectorTypes.DOWN, ch)
    
    if dynamic:
      self.file_path = path
      self._timer_(self.enable_dynamic_topology, nodetype, channeltype)

  def _timer_(self, function, *args):
    stopped = threading.Event()

    def loop():
        while not stopped.wait(random() * 10): # the first call is in `interval` secs
                function(*args)

    threading.Thread(target=loop).start()    
    return stopped.set

  def enable_dynamic_topology(self,
      nodetype,
      channeltype,
      new_node: float = 0.6, # the probability of creating new node
      delete_node: float = 0.4, # the probability of deletion of current nodes
      new_edge: float = 0.45, # the probability of creating new edge
      delete_edge: float = 0.45, # the probability of deletion new edge 
      ) -> None:

      flags = {
        "new_node": random() <= new_node,
        "delete_node": random() <= delete_node,
        "new_edge": random() <= new_edge,
        "delete_edge": random() <= delete_edge
      }

      if flags["new_node"]:
        u = max(self.nodes) + 1
        v = sample(sorted(self.nodes), 1)[0]
        self.G.add_node(u)
        self.G.add_edge(u, v)
        cc = nodetype(nodetype.__name__, u, u)
        self.nodes[u] = cc

        ch = channeltype(channeltype.__name__, str(u) + "-" + str(v))
        self.channels[(u,v)] = ch
        self.nodes[u].connect_me_to_channel(ConnectorTypes.DOWN, ch)
        self.nodes[v].connect_me_to_channel(ConnectorTypes.DOWN, ch)

      # if flags["delete_node"]:
      #   self.G.remove_node(sample(self.nodes, 1)[0])

      if flags["new_edge"]:
        all_possible_edges = list(itertools.combinations(self.nodes.keys(), 2))
        for i in self.channels.keys():
          all_possible_edges.remove((min(i),max(i)))
        u, v = sample(all_possible_edges, 1)[0]

        self.G.add_edge(u, v)
        ch = channeltype(channeltype.__name__, str(u) + "-" + str(v))
        self.channels[(u,v)] = ch
        self.nodes[u].connect_me_to_channel(ConnectorTypes.DOWN, ch)
        self.nodes[v].connect_me_to_channel(ConnectorTypes.DOWN, ch)

      if flags["delete_edge"]:
         u, v = sample(self.channels.keys(), 1)[0]
         node_u = self.nodes[u]
         node_v = self.nodes[v]
         self.G.remove_edge(u, v)
         ch = self.channels[(u,v)]
         node_u.remove_connection_to_channel_from_me(ConnectorTypes.DOWN, ch)
         node_v.remove_connection_to_channel_from_me(ConnectorTypes.DOWN, ch)

      with open(self.file_path, "a") as f:
        f.write(str(time.time_ns() / 1000) + '\n')
        f.write(str(self.G.nodes) + '\n')
        f.write(str(self.G.edges) + '\n')
        f.write('\n')

# TODO: construct_from_graph_peterson and construct_from_graph_bakery will be removed... Does not follow the AHC style..
  def construct_from_graph_peterson(self, G: nx.Graph, nodetype, channeltype):
    self.G = G
    nodes = list(G.nodes)
    edges = list(G.edges)

    nodes = nodes[::-1]
    edges = edges[::-1]

    for i in nodes:
      cc = nodetype(nodetype.__name__, i, i)
      self.nodes[i] = cc
    for k in edges:
      ch = channeltype(channeltype.__name__, str(k[0]) + "-" + str(k[1]))
      self.channels[k] = ch
      self.nodes[k[0]].connect_me_to_channel(ConnectorTypes.DOWN, ch)
      self.nodes[k[1]].connect_me_to_channel(ConnectorTypes.DOWN, ch)

  def construct_from_graph_bakery(self, G: nx.Graph, nodetype, channeltype):
    self.G = G
    nodes = list(G.nodes)
    edges = list(G.edges)

    nodes = nodes[::-1]
    edges = edges[::-1]

    for i in nodes:
      cc = nodetype(nodetype.__name__, i, i)
      self.nodes[i] = cc
    for k in edges:
      ch = channeltype(channeltype.__name__, str(k[0]) + "-" + str(k[1]))
      self.channels[k] = ch
      self.nodes[k[0]].connect_me_to_channel(ConnectorTypes.DOWN, ch)
      self.nodes[k[1]].connect_me_to_channel(ConnectorTypes.DOWN, ch)

  def construct_single_node(self, nodetype, instancenumber):
    self.singlenode = nodetype(nodetype.__name__, instancenumber)
    self.G = nx.Graph()
    self.G.add_nodes_from([0])
    self.nodes[0] = self.singlenode

  def construct_sender_receiver(self, sendertype, receivertype, channeltype):
    self.sender = sendertype(sendertype.__name__, 0)
    self.receiver = receivertype(receivertype.__name__, 1)
    ch = channeltype(channeltype.__name__, "0-1")
    self.G = nx.Graph()
    self.G.add_nodes_from([0, 1])
    self.G.add_edges_from([(0, 1)])
    self.nodes[self.sender.componentinstancenumber] = self.sender
    self.nodes[self.sender.componentinstancenumber] = self.receiver
    self.channels[ch.componentinstancenumber] = ch
    self.sender.connect_me_to_channel(ConnectorTypes.DOWN, ch)
    self.receiver.connect_me_to_channel(ConnectorTypes.DOWN, ch)

  def allpairs_shortest_path(self):
    return dict(nx.all_pairs_shortest_path(self.G))

  def shortest_path_to_all(self, myid):
    path = dict(nx.all_pairs_shortest_path(self.G))
    nodecnt = len(self.G.nodes)
    for i in range(nodecnt):
      print(path[myid][i])

  def start(self):
    # registry.printComponents()
    N = len(self.G.nodes)
    self.compute_forwarding_table()
    self.nodecolors = ['b'] * N
    # self.nodepos = nx.drawing.spring_layout(self.G)
    self.lock = Lock()
    ComponentRegistry().init()

  def compute_forwarding_table(self):
    # N = len(self.G.nodes)
    self.ForwardingTable = dict(nx.all_pairs_shortest_path(self.G))
    # print(f"There are {N} nodes")
    # for i in range(N):
      # for j in range(N):
        # try:
          # mypath = path[i][j]
          # print(f"{i}to{j} path = {path[i][j]} nexthop = {path[i][j][1]}")
          # self.ForwardingTable[i][j] = path[i][j][1]

          # print(f"{i}to{j}path = NONE")
          # self.ForwardingTable[i][j] = inf  # No paths
        # except IndexError:
          # print(f"{i}to{j} nexthop = NONE")
          # self.ForwardingTable[i][j] = i  # There is a path but length = 1 (self)

  # all-seeing eye routing table contruction
  def print_forwarding_table(self):
    registry.print_components()
    print('\n'.join([''.join(['{:4}'.format(item) for item in row])
                     for row in list(self.ForwardingTable.values())]))

  # returns the all-seeing eye routing based next hop id
  def get_next_hop(self, fromId, toId):
    try:
      retval = self.ForwardingTable[fromId][toId]
      return retval[1]
    except KeyError:
      return inf
    except IndexError:
      return fromId

  # Returns the list of neighbors of a node
  def get_neighbors(self, nodeId):
    return sorted([neighbor for neighbor in self.G.neighbors(nodeId)])

  def get_predecessors(self, nodeId):
    return sorted([neighbor for neighbor in self.G.predecessors(nodeId)])

  def get_successors(self, nodeId):
    return sorted([neighbor for neighbor in self.G.neighbors(nodeId)])

  # Returns the list of neighbors of a node
  def get_neighbor_count(self, nodeId):
    # return len([neighbor for neighbor in self.G.neighbors(nodeId)])
    return self.G.degree[nodeId]

  def plot(self):
    # self.lock.acquire()
    # nx.draw(self.G, self.nodepos, node_color=self.nodecolors, with_labels=True, font_weight='bold')
    nx.draw(self.G, self.nodepos, node_color=self.nodecolors, with_labels=True, font_weight='bold')
    # plt.draw()
    print(self.nodecolors)
    # self.lock.release()

  def get_random_node(self):
    return self.nodes[sample(self.G.nodes(), 1)[0]]

@singleton
class FramerObjects():
    framerobjects = {}
    ahcuhdubjects = {}
    def add_framer(self, id, obj):
        self.framerobjects[id] = obj
    
    def get_framer_by_id(self, id):
        return self.framerobjects[id]

    def add_ahcuhd(self, id, obj):
        self.ahcuhdubjects[id] = obj
    
    def get_ahcuhd_by_id(self, id):
        return self.ahcuhdubjects[id]



framers = FramerObjects()





'''
        This module is designed for the implementation of Distance Sequenced Distance Vector routing algorithm. In its paper,
    some parts are ambiguous. Therefore, some parts are decided by the developer itself.
'''

__author__ = "Bahadır Kisbet"
__contact__ = "bahadirkisbet@gmail.com"
__copyright__ = "Copyright 2021, WINSLAB"
__credits__ = ["Bahadır Kisbet"]
__date__ = "2021/04/07"  # will be updated later
__deprecated__ = False
__email__ = "bahadirkisbet@gmail.com"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Development"
__version__ = "0.0.1"

import sys
import os

sys.path.insert(0, os.getcwd())


import matplotlib.pyplot as plt
from enum import Enum
import numpy as np
import networkx as nx
from ahc.Ahc import (ComponentModel, Event, EventTypes, GenericMessage,
                 GenericMessageHeader, GenericMessagePayload, Topology)
import threading
from ahc.Channels.Channels import P2PFIFOPerfectChannel
from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer
from ahc.Ahc import ConnectorTypes
from tabulate import tabulate
from time import sleep, time
import copy

# @TODO: Implement dynamically changing topology
class DSDVMessageTypes(Enum):
    '''
        INCREMENTAL_DUMP: It is a message which contains only the changes since the last broadcast
        FULL_DUMP: It is a message which contains all routing table. It is sent under specific conditions
    '''
    INCREMENTAL = "INCREMENTAL_DUMP"
    FULLDUMP = "FULL_DUMP"

class DSDVMessageHeader(GenericMessageHeader):
    def __init__(self, messagetype, messagefrom, messageto, nexthop=float('inf'), interfaceid=float('inf'), messagesource=-1, hopcount=0, sequencenumber=-1):
        super().__init__(messagetype, messagefrom, messageto, nexthop=nexthop, interfaceid=interfaceid, sequencenumber=sequencenumber)
        self.messagesource = messagesource
        self.hopcount = hopcount


class DSDVNode(ComponentModel): # It is an application layer
    def __init__(self,
        componentname: str,
        componentinstancenumber: str, 
        interval: float = 3.0,
        timeout: float = 7.0,
        expiration_time: int = 3, # if a node does not send a broadcast in interval, its expiration will increase by one. If it reaches to three, it will be accepted as broken link.
        verbose: bool = True) -> None:

        super().__init__(componentname=componentname, componentinstancenumber=componentinstancenumber)

        self.sequence_num = 0
        self.sequence_num_lock = threading.Lock()

        # every entry is -> destination: (nextHop, hopCount, seqnum)
        self.routing_table = {componentinstancenumber: (componentinstancenumber, 0, 0)} 
        self.lock = threading.Lock()

        self.incremental_table = copy.deepcopy(self.routing_table)
        self.incremental_table_lock = threading.Lock()

        self.cache = dict()
        self.cache_lock = threading.Lock()

        self.verbose = verbose
        self.log_table_path = f"ahc/Routing/DSDV/logs/graph2/log_table_{self.componentinstancenumber}.txt"
        self.log_throughput_path = f"ahc/Routing/DSDV/logs/graph2/log_throughput_{self.componentinstancenumber}.txt"
        
        self.expiration_dict = dict() # it keeps track of which neighbours havent send an update for expiration time
        self.expiration_lock = threading.Lock()

        self.timeout = timeout
        self._timer_(2, self._broadcast )
        self._timer_(3, self.print_routing_table)
        # self._timer_(interval, self.updateMsgToNeighbours)
        #threading.Timer(3.0, self.updateMsgToNeighbours).start()

    def _timer_(self, interval, function, *args):
        stopped = threading.Event()

        def loop():
            while not stopped.wait(interval): # the first call is in `interval` secs
                    function(*args)

        threading.Thread(target=loop).start()    
        return stopped.set
    
    def on_message_from_bottom(self, eventobj: Event):
        super().on_message_from_bottom(eventobj)

        header: GenericMessageHeader = eventobj.eventcontent.header
        payload: GenericMessagePayload = eventobj.eventcontent.payload

        if header.messagetype == DSDVMessageTypes.INCREMENTAL or header.messagetype == DSDVMessageTypes.FULLDUMP:
            self._handle_incremental(header, payload)

    def __increase_timeouts__(self) -> None:
        with self.expiration_lock:
            for entry in self.expiration_dict:
                self.expiration_dict[entry] += 1
   
    def print_routing_table(self) -> None:
        with self.lock, open(self.log_table_path, "a+") as f:
            headers = ["DESTINATION", "NEXT_HOP", "HOP_COUNT", "SEQ_NUM"]
            data = [ ",".join([str(entry)] + list(map(str,self.routing_table[entry]))) for entry in self.routing_table ]
            f.write(f"{time()}\n")
            f.write(str(self.componentinstancenumber) + "\n")
            f.write("\n".join(data))
            f.write('\n')

    ### NEW ###
    def _prepare_dump(self, to: int) -> GenericMessage:
        msg = None
        with self.lock:
            changes = dict()
            for element in self.routing_table:
                if element in self.incremental_table:
                    if self.routing_table[element] != self.incremental_table[element]:
                        changes[element] = self.routing_table[element]
                else:
                    changes[element] = self.routing_table[element]
            
            msg_type = DSDVMessageTypes.INCREMENTAL
            msg_payload = changes
            if len(changes) > len(self.routing_table) // 2:
                msg_type = DSDVMessageTypes.FULLDUMP
                msg_payload = self.routing_table

            header = DSDVMessageHeader(
                messagetype=msg_type,
                messagefrom=self.componentinstancenumber,
                messagesource=self.componentinstancenumber,
                hopcount=1,
                messageto=to,
                interfaceid=f"{self.componentinstancenumber}-{to}",
                nexthop=self.componentinstancenumber,
                sequencenumber=self.sequence_num
            )
            payload = GenericMessagePayload(msg_payload)
            msg = GenericMessage(header, payload)
        return msg

    def _prepare_message(self, to: int) -> GenericMessage:
        return self._prepare_dump(to)

    def _broadcast(self, msg: GenericMessage = None, is_mine = True) -> None:
        if is_mine:
            with self.sequence_num_lock, self.lock:
                self.sequence_num += 1
                self.routing_table[self.componentinstancenumber] = (self.componentinstancenumber, 0, self.sequence_num)
        
        packet = msg
        for neighbour in Topology().get_neighbors(self.componentinstancenumber):

            if msg == None:
                packet = self._prepare_message(neighbour)

            with open(self.log_throughput_path, "a+") as f:
                f.write(f"{time()}\t{sys.getsizeof(packet)}\n")
            self.send_down(Event(self, EventTypes.MFRT, packet))

    def _handle_incremental(self, header: DSDVMessageHeader, payload: GenericMessagePayload) -> None:
        with self.lock, self.cache_lock, self.expiration_lock:
            if header.messagefrom not in self.cache or self.cache[header.messagefrom] < header.sequencenumber:
                # The message is new
                self.cache[header.messagesource] = header.sequencenumber
                changes = payload.messagepayload
                
                others = dict() # it is a payload to send the broadcast to other nodes

                if header.messagesource not in self.routing_table:
                    self.routing_table[header.messagesource] = (header.messagefrom, header.hopcount, header.sequencenumber)

                if header.messagesource in self.expiration_dict: # it is cached before
                    curr_time = time()
                    self.expiration_dict[header.messagesource] = curr_time + self.timeout
                    for src in self.expiration_dict:
                        if self.expiration_dict[src] < curr_time:
                            self.routing_table[src] = (
                                self.routing_table[src][0],
                                float('inf'),
                                self.routing_table[src][0] + 1 )
                else:
                    self.expiration_dict[header.messagesource] = time() + self.timeout
                

                for change in changes:
                    if change in self.routing_table: # if it can be reachable from the source
                        #print(self.componentinstancenumber, change, self.routing_table[change][1], self.routing_table[header.messagesource][1] + changes[change][1])

                        if self.routing_table[change][1] >= self.routing_table[header.messagesource][1] + changes[change][1]:
                            # Bell-man Ford
                            if changes[change][-1] >= self.routing_table[change][-1]: # if the seqnum is more recent
                                self.routing_table[change] = (
                                    header.messagefrom,
                                    self.routing_table[header.messagesource][1] + changes[change][1],
                                    changes[change][-1])
                    else: # add to the routing table adding one hop count more
                        self.routing_table[change] = (
                            header.messagefrom,
                            changes[change][1] + 1,
                            changes[change][-1]
                        )
                    
                    others[change] = (
                            header.messagefrom,
                            changes[change][1],
                            changes[change][-1]
                        )
                header.messagefrom = self.componentinstancenumber
                header.hopcount += 1
                payload = GenericMessagePayload(others)
                #print(len(others), sys.getsizeof(others),sys.getsizeof(payload))
                self._broadcast(GenericMessage(header, payload), is_mine=False) # forward the packet
            else: # drop it
                pass


class AdHocNode(ComponentModel):

    def on_init(self, eventobj: Event):
        print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def __init__(self, componentname, componentid):
        # SUBCOMPONENTS
        self.applicationlayer   = DSDVNode("DSDVNode", componentid, 3.0)
        self.networklayer       = AllSeingEyeNetworkLayer("NetworkLayer", componentid)
        self.linklayer          = LinkLayer("LinkLayer", componentid)
        # self.failuredetect = GenericFailureDetector("FailureDetector", componentid)

        # CONNECTIONS AMONG SUBCOMPONENTS
        self.applicationlayer.connect_me_to_component(ConnectorTypes.DOWN, self.networklayer)
        self.networklayer.connect_me_to_component(ConnectorTypes.UP, self.applicationlayer)

        self.networklayer.connect_me_to_component(ConnectorTypes.DOWN, self.linklayer)
        self.linklayer.connect_me_to_component(ConnectorTypes.UP, self.networklayer)

        # Connect the bottom component to the composite component....
        self.linklayer.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.linklayer)

        super().__init__(componentname, componentid)


def main():
    # G = nx.Graph()
    # G.add_nodes_from([1, 2])
    # G.add_edges_from([(1, 2)])
    # nx.draw(G, with_labels=True, font_weight='bold')
    # plt.draw()

    #G = nx.random_geometric_graph(4, 0.5)
    # G = nx.Graph()
    # G.add_nodes_from([1,2,3,4,5,6,7,8,9,10,11,12])
    # G.add_edges_from([
    #     (1,2),
    #     (2,5),
    #     (3,4),
    #     (3,5),
    #     (4,5),
    #     (5,6),
    #     (6,7),
    #     (5,7),
    #     (5,8),
    #     (6,12),
    #     (7,10),
    #     (7,8),
    #     (8,9),
    #     (9,11),
    #     (10,11),
    #     (11,12)
    # ])
    G = nx.Graph()
    G.add_nodes_from([1,2,3,4])
    G.add_edges_from([
        (1,2),
        (1,3),
        (2,3),
        (2,4)
    ])

    nx.draw(G, with_labels=True, font_weight='bold')
    plt.draw()
    topo = Topology()
    topo.construct_from_graph(G, DSDVNode, P2PFIFOPerfectChannel, dynamic=False, path="/Users/bahadirkisbet/Desktop/projects/ahc/Routing/DSDV/logs/topology.txt")
    topo.start()
    sleep(15)
    exit(0)
    plt.show()
    #plt.show()



if __name__ == "__main__":
    main()



from ahc.Ahc import Topology
from ahc.Channels.Channels import P2PFIFOPerfectChannel
import networkx as nx
from ahc.Routing.GSR.GSRTestingNodeComponent import GSRTestingNode
from ahc.Routing.GSR.RoutingGSRComponent import RoutingGSRComponent


# undirected graph
def draw_random_graph(n):
    k = True
    g_random = None
    while k:
        k = False
        g_random = nx.gnp_random_graph(n, 0.3)
        if not nx.is_connected(g_random):
            k = True
    return g_random


# undirected graph
NODE_COUNT = 10

if __name__ == "__main__":
    graph = draw_random_graph(NODE_COUNT)  # nx.Graph()
    print(graph.edges)
    # graph.add_edges_from(edges)

    node_list = graph.nodes

    print(RoutingGSRComponent.__name__)

    topology = Topology()
    topology.construct_from_graph(graph, GSRTestingNode, P2PFIFOPerfectChannel)
    topology.start()

    while True:
        pass



import binascii
import hashlib
import sys
import threading
import time
from copy import deepcopy
from enum import Enum

import networkx as nx
from random import choice

import matplotlib.pyplot as plt

from ahc.Channels.Channels import Channel
from ahc.Ahc import ComponentRegistry, Topology, GenericMessageHeader, GenericMessage, EventTypes, Event, ComponentModel

PATH_OF_DATA = "data"
sys.setrecursionlimit(1500)
def transaction_generator(numberofNodes,txn_count):
    for n in range(0,numberofNodes):
        filename =  f"{PATH_OF_DATA}/transaction{n}.txt"
        file = open(filename,"w")
        for i in range(0, txn_count):
            rnd_hex=''.join([str(y) for x in range(64) for y in choice('0123456789abcdef')])
            file.write(rnd_hex + "\n")
        file.close()
def hash_double(firstTxHash, secondTxHash):
    unhexReverseFirst = binascii.unhexlify(firstTxHash)[::-1]
    unhexReverseSecond = binascii.unhexlify(secondTxHash)[::-1]
    concatInputs = unhexReverseFirst + unhexReverseSecond
    firstHashInputs = hashlib.sha256(concatInputs).digest()
    finalHashInputs = hashlib.sha256(firstHashInputs).digest()
    return binascii.hexlify(finalHashInputs[::-1])

def merkle_root_calculator(hashList):
    if len(hashList) == 1:
        return hashList[0].decode('utf-8')
    hashes = []
    for i in range(0, len(hashList) - 1, 2):
        hashes.append(hash_double(hashList[i], hashList[i + 1]))
    if len(hashList) % 2 == 1:
        hashes.append( hashList[-1])
    return merkle_root_calculator(hashes)

class Block:
    def __init__(self, hashPrevHeader, merkleroot, timestamp, transactions, blockhash, nonce , height = 0):
        self.hashPrevHeader = hashPrevHeader
        self.merkleRoot = merkleroot
        self.timestamp = timestamp
        self.transactions = transactions
        self.nonce = nonce
        self.hashBlockHeader = blockhash
        self.height = height

class Blockchain:
    def __init__(self, difficulty=4):
        self.blocks = []
        self.difficulty = difficulty
        self.generate_genesis_block()


    def generate_genesis_block(self):
        genesis_block = Block("0", "0",time.time(), [], "0", 0, 0)
        self.blocks.append(genesis_block)

    def get_last_block(self):
        return self.blocks[-1]

class NkMessageTypes(Enum):
    TXN = "transaction"
    VLD = "validation"
    QUERY_LC = "querylongestchain"
    RW_LC = "responsewithlongestchain"


class NkComponent(ComponentModel):
    def __init__(self, componentname, componentinstancenumber, difficulty=5):
        super().__init__(componentname, componentinstancenumber)
        self.blockchain = Blockchain()
        self.txnPool = []
        self.txnLog = []
        self.difficulty = difficulty
        self.isMiner = True

        # To test components with different registeries
        self.speedDown = 1

        self.lock = threading.Lock()
        minerThread = threading.Thread(target=self.mine_next_block)
        minerThread.daemon = True
        minerThread.start()

    def on_init(self, eventobj: Event):

        with open(f"data/transaction{self.componentinstancenumber}.txt", "r") as txns:
            txn_list = txns.read()
        self.txnPool = txn_list.splitlines()
        self.passNextBlock = True

    def print_chain(self):
        print(f"{self.componentname}-{self.componentinstancenumber} blockchain is : ")
        for b in self.blockchain.blocks:
            print(f"{b.hashPrevHeader}-{b.hashBlockHeader}\n")

    def on_message_from_bottom(self, eventobj: Event):
        msg = eventobj.eventcontent
        msgType = msg.header.messagetype
        msgPayload = msg.payload
        msgfrom = msg.header.messagefrom
        msgto = msg.header.messageto
        '''
        Three types of message could be transmitted:
        1) Transaction Messages : A new transaction from client is added to transaction pool and 
           the node broadcast that to nearby nodes (probably miners)
        2) Validation Message : Recently mined  block is recieved for validation from other nodes and local blockchain is 
        updated if it obeys to rules
        3) Blockchain update message : Periodically a chain queries the blockchains from neighbour nodes, if their blockchain
        is more up to date and valid, replace current chain with more recent one  and add transactions stored in current 
        blockchain to pool which doesnt included in new chain

        '''
        self.lock.acquire()
        if msgType == NkMessageTypes.TXN:
            if msgPayload in self.txnPool:
                pass
            else:
                self.txnPool.append(msgPayload)
                nk_header = GenericMessageHeader(NkMessageTypes.TXN, self, None)
                nk_payload = msgPayload
                nk_eventContent = GenericMessage(nk_header, nk_payload)
                nk_event = Event(self, EventTypes.MFRT, nk_eventContent)
                self.send_down(nk_event)

        elif msgType == NkMessageTypes.VLD:
            if self.is_valid(msgPayload) and self.is_prev_correct(msgPayload):
                self.txnPool = [txn for txn in self.txnPool if not txn in msgPayload.transactions]
                self.blockchain.blocks.append(msgPayload)
                nk_header = GenericMessageHeader(NkMessageTypes.VLD, self, None)
                nk_payload = msgPayload
                nk_eventContent = GenericMessage(nk_header, nk_payload)
                nk_event = Event(self, EventTypes.MFRT, nk_eventContent)
                self.send_down(nk_event)
                print(f"Component {self.componentinstancenumber}  validated a new block: "
                      f"txn count= {len(self.txnPool)} - block count= {len(self.blockchain.blocks)}\n"
                      f"{self.blockchain.get_last_block().hashPrevHeader}")
                self.passNextBlock = True

            elif msgPayload.height > self.blockchain.get_last_block().height:
                print(f"Component {self.componentinstancenumber}  the message block height is higher ")
                self.search_longest_chain()

        elif msgType == NkMessageTypes.QUERY_LC:
            # response to sender
            nk_header = GenericMessageHeader(NkMessageTypes.RW_LC, self, msgfrom)
            nk_payload = deepcopy(self.blockchain)
            nk_eventContent = GenericMessage(nk_header, nk_payload)
            nk_event = Event(self, EventTypes.MFRT, nk_eventContent)
            self.send_down(nk_event)

        elif msgType == NkMessageTypes.RW_LC and msgto is self:
            if msgPayload.get_last_block().height > self.blockchain.get_last_block().height:
                nTxn = []
                self.txnPool.extend(self.txnLog)  # Restore back all txn mined before to txn pool
                self.txnLog.clear() # Not necessary but memory overload causes crashes
                for block in msgPayload.blocks:
                    nTxn.extend(block.transactions)
                self.txnPool = [txn for txn in self.txnPool if txn not in nTxn]  # Eliminate the ones in new ledger
                self.blockchain = msgPayload
                self.passNextBlock = False
                print(f"Component {self.componentinstancenumber} received a longer chain: "
                      f"txn count= {len(self.txnPool)} - block count= {len(self.blockchain.blocks)}\n"
                      f"{self.blockchain.get_last_block().hashPrevHeader}")

            else:
                pass
        self.lock.release()

    def is_valid(self, inBlock: Block):
        prvHeader = inBlock.hashPrevHeader
        mrklRoot = inBlock.merkleRoot
        nonce = inBlock.nonce
        hdr = (mrklRoot + str(nonce) + prvHeader).encode('utf-8')
        blockHash = hashlib.sha256(hdr).hexdigest()
        return blockHash.startswith("0" * self.difficulty)

    def is_prev_correct(self, inBlock: Block):
        currBlockHash = self.blockchain.blocks[-1].hashBlockHeader
        inBlockPrvHash = inBlock.hashPrevHeader
        return currBlockHash == inBlockPrvHash

    def mine_next_block(self, minTxnPackage=10):
        while self.isMiner:
            if minTxnPackage <= len(self.txnPool):
                txnPending = self.txnPool[:minTxnPackage]
                lastBlock = self.blockchain.get_last_block()
                hashPrevHeader = lastBlock.hashBlockHeader
                height = lastBlock.height + 1
                merkleRoot = merkle_root_calculator(txnPending)
                nonce = 0
                isMined = False
                self.passNextBlock = False
                time.sleep(self.speedDown)
                while not self.passNextBlock and not isMined:
                    hdr = (merkleRoot + str(nonce) + hashPrevHeader).encode('utf-8')
                    blockHash = hashlib.sha256(hdr).hexdigest()
                    if blockHash.startswith("0" * self.difficulty):
                        self.lock.acquire()
                        self.txnPool = [txn for txn in self.txnPool if txn not in txnPending]
                        self.txnLog.extend(txnPending)
                        currTime = time.localtime()
                        newBlock = Block(hashPrevHeader, merkleRoot, currTime, txnPending, blockHash, nonce, height)
                        self.blockchain.blocks.append(newBlock)

                        nkHeader = GenericMessageHeader(NkMessageTypes.VLD, self, None)
                        nkPayload = newBlock
                        nkEventContent = GenericMessage(nkHeader, nkPayload)
                        nkEvent = Event(self, EventTypes.MFRT, nkEventContent)
                        self.send_down(nkEvent)

                        isMined = True
                        print(f"Component {self.componentinstancenumber}  mined a new block: "
                              f"txn count= {len(self.txnPool)} - block count= {len(self.blockchain.blocks)}\n"
                              f"{self.blockchain.get_last_block().hashPrevHeader}")
                        self.lock.release()
                    else:
                        nonce = nonce + 1

    def search_longest_chain(self):
        self.passNextBlock = True
        print(f"Component {self.componentinstancenumber} is asking for longer chain")
        nk_payload = "asking for longer chain"
        nk_header = GenericMessageHeader(NkMessageTypes.QUERY_LC, self, None)
        nk_eventContent = GenericMessage(nk_header, nk_payload)
        nk_event = Event(self, EventTypes.MFRT, nk_eventContent)
        self.send_down(nk_event)

def main():
    number_of_nodes = 5
    number_of_txn = 50
    G = nx.erdos_renyi_graph(number_of_nodes, 0.4)
    transaction_generator(number_of_nodes,number_of_txn)
    topo = Topology()
    topo.construct_from_graph(G, NkComponent, Channel)

    ComponentRegistry().print_components()
    topo.start()
    nx.draw(G, with_labels=True, font_weight='bold')
    plt.show()

if __name__ == '__main__':
    main()
    while True: pass


from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes, ComponentRegistry, Lock, Thread, Topology
import time
import numpy as np

class FredericksonAlgorithmAdvancedComponent(ComponentModel):
    def __init__(self, componentname, componentid):
        super(FredericksonAlgorithmAdvancedComponent, self).__init__(componentname, componentid)
        self.queue_lock = Lock()
        self.message_queue = []
        self.l_parameter = 1
        if self.componentinstancenumber == 0:
            self.is_initiator = True
        else:
            self.is_initiator = False

    def on_init(self, eventobj: Event):
        super(FredericksonAlgorithmAdvancedComponent, self).on_init(eventobj)
        if not self.is_initiator:
            thread = Thread(target=self.job, args=[45, 54, 123])
            thread.start()

    def on_message_from_bottom(self, eventobj: Event):
        message_destination = eventobj.eventcontent.header.messageto.split("-")[0]
        print(f"{self.componentinstancenumber} received {message_destination}")
        if message_destination == FredericksonAlgorithmAdvancedComponent.__name__: # process only the messages targeted to this component...
            message_source_id = eventobj.eventcontent.header.messagefrom.split("-")[1]
            message_type = eventobj.eventcontent.header.messagetype
            content = eventobj.eventcontent.payload
            if message_type == "EXPLORE" or message_type == "FORWARD" or message_type=="REVERSE":
                self.queue_lock.acquire() # protect message_queue, both component thread and Toueg thread are trying to access data
                self.message_queue.append((int(message_source_id), message_type, content))
                self.queue_lock.release()

    def on_message_from_peer(self, eventobj: Event):
        message_header = eventobj.eventcontent.header
        message_target = eventobj.eventcontent.header.messageto.split("-")[0]
        if message_target == FredericksonAlgorithmAdvancedComponent.__name__:
            if self.is_initiator:
                if message_header.messagetype == "INITIATEBFSCONSTRUCTION":
                    thread = Thread(target=self.job, args=[45, 54, 123])
                    thread.start()

    def job(self, *arg):
        self.neighbors = Topology().get_neighbors(self.componentinstancenumber)  # retrieve all neighbor ids...
        self.neighbor_weights = {a: 1 for a in self.neighbors}  # for the time being each edge weight is 1...
        if self.componentinstancenumber == 0:
            self.is_initiator = True
        else:
            self.is_initiator = False

        tree = self.FredericksonAlgorithmAdvanced()
        def getPaths(data):
            if len(data) == 0:
                return [[]]
            if len(data) == 1:
                ret = getPaths(data[list(data.keys())[0]])
                for i in ret:
                    i.insert(0, list(data.keys())[0])
                return ret
            if len(data) > 1:
                ret = getPaths(data[list(data.keys())[0]])
                for kl in range(len(ret)):
                    ret[kl].insert(0, list(data.keys())[0])

                return ret + getPaths({a: data[a] for a in list(data.keys()) if a != list(data.keys())[0]})

        message_payload= getPaths(tree)

        message_header = GenericMessageHeader("BFSTREECONSTRUCTED",
                                              self.componentname + "-" + str(self.componentinstancenumber),
                                              "Coordinator-" + str(self.componentinstancenumber))
        message = GenericMessage(message_header, message_payload)
        event = Event(self, EventTypes.MFRP, message)
        self.send_peer(event)
        print("Info sent to Coordinator")

    def FredericksonAlgorithmAdvanced(self):
        self.process_id = self.componentinstancenumber
        self.positively_responded_nodes = []

        self.level_u = np.inf
        self.neighbor_level_u = {}
        self.parent_u = None
        self.children_u = {}
        self.expectedreplies = {}
        for neighbor in self.neighbors:
            self.neighbor_level_u[neighbor] = np.inf
            self.expectedreplies[neighbor] = 0

        bvalue_u = False
        sendreverse_u = True

        if self.is_initiator:
            self.level_u = 0
            k = 0
            for n in self.neighbors:
                if not n in self.children_u:
                    self.children_u[n] = []
                self.sendMessageToNeighbor(n, "EXPLORE", (k + 1, self.l_parameter))
                self.expectedreplies[n] = 1

        while True:
            new_message = self.waitNewMessage()
            sender, message_type, f = new_message
            search_depth = f

            if message_type == "FORWARD":
                bvalue_u = False
                for n in self.neighbors:
                    self.expectedreplies[n] = 0
                if self.level_u < f:
                    message_count = 0
                    for c in self.positively_responded_nodes:  # list(set(self.children_u)):
                        self.sendMessageToNeighbor(c, "FORWARD", f)
                        self.expectedreplies[c] = 1
                        message_count += 1
                    self.positively_responded_nodes = []


                if self.level_u == f:
                    message_count = 0
                    for n in self.neighbors:
                        if self.neighbor_level_u[n] != f - 1:
                            self.sendMessageToNeighbor(n, "EXPLORE", (f + 1, self.l_parameter))
                            self.expectedreplies[n] = 1
                            message_count += 1
                    if message_count == 0:
                        self.message_queue.append((-1, "REVERSE", (False, self.children_u)))
                        print(f"**********{self.process_id} sends loop back itsel... ")

            elif message_type == "EXPLORE":
                f = search_depth[0]
                m = search_depth[1]

                if self.neighbor_level_u[sender] != f - 1:
                    self.neighbor_level_u[sender] = f - 1

                if self.level_u > f:
                    bvalue_u = True
                    if not sendreverse_u:
                        self.sendMessageToNeighbor(self.parent_u, "REVERSE", (False, self.children_u))
                        sendreverse_u = True

                    self.parent_u = sender
                    self.level_u = f

                    self.children_u = {}
                    if m > 1:
                        sendreverse_u = False
                        message_count = 0
                        for n in self.neighbors:
                            if n != self.parent_u:
                                self.sendMessageToNeighbor(n, "EXPLORE", (f + 1, m - 1))
                                self.expectedreplies[n] += 1
                                message_count += 1

                    else:
                        self.sendMessageToNeighbor(sender, "REVERSE", (True, self.children_u))
                elif self.level_u == f or self.level_u == f - 1:
                    if sender in self.children_u:
                        del self.children_u[sender]
                    if sender in self.positively_responded_nodes:
                        self.positively_responded_nodes.remove(sender)

                    self.sendMessageToNeighbor(sender, "REVERSE", (False, self.children_u))
                elif self.level_u < f - 1:
                    self.sendMessageToNeighbor(sender, "REVERSE", (False, self.children_u))

            elif message_type == "REVERSE":
                b = f[0]
                if sender in self.expectedreplies:
                    self.expectedreplies[sender] -= 1
                if sender in self.neighbor_level_u and self.neighbor_level_u[sender] <= self.level_u:
                    b = False
                if b == True:
                    if sender not in self.positively_responded_nodes:
                        self.positively_responded_nodes.append(sender)
                    if self.process_id != sender:
                        self.children_u[sender] = f[1]
                    bvalue_u = True

                all_responded = True
                for i in self.expectedreplies:
                    if self.expectedreplies[i] != 0:
                        all_responded = False
                        break

                if all_responded == True:
                    if self.parent_u != None:
                        self.sendMessageToNeighbor(self.parent_u, "REVERSE", (bvalue_u, self.children_u))
                        sendreverse_u = False
                    elif bvalue_u == True:
                        bvalue_u = False
                        k = k + 1
                        message_count = 0
                        for c in self.positively_responded_nodes:  # list(set(self.children_u)):
                            self.sendMessageToNeighbor(c, "FORWARD", k)
                            self.expectedreplies[c] = 1
                            message_count += 1
                        self.positively_responded_nodes = []

                        if message_count == 0:
                            print("BFS Completed....")
                            print(f"**********{self.process_id} has problem {new_message}")
                            print(self.children_u)
                            break
                    else:
                        print("BFS Completed....")
                        break

        print(self.children_u)
        return {self.componentinstancenumber: self.children_u}

    def sendMessageToNeighbor(self, neighbor_id, message_type, message):
        print(f"{self.componentinstancenumber} sends {message_type} message to neighbor {neighbor_id}")
        message_header = GenericMessageHeader(message_type, FredericksonAlgorithmAdvancedComponent.__name__+"-"+str(self.componentinstancenumber),
        FredericksonAlgorithmAdvancedComponent.__name__+"-"+str(neighbor_id), interfaceid=str(self.componentinstancenumber)+"-"+str(neighbor_id))
        mess_ = GenericMessage(message_header, message)

        event = Event(self, EventTypes.MFRT, mess_)
        self.send_down(event)

    def waitNewMessage(self):
        self.queue_lock.acquire()
        if len(self.message_queue) > 0:
            message = self.message_queue.pop()
            sender = message[0]
            message_type = message[1]
            last_part = message[2]
            self.queue_lock.release()
            return (sender, message_type, last_part)
        else:
            self.queue_lock.release()
            return None, None, None


GSR_ROUTER_NAME = "GSR_ROUTER_COMPONENT"
GSR_COORDINATOR_NAME = "GSR_COORDINATOR_COMPONENT"
GSR_APPLICATION_NAME = "GSR_APPLICATION_COMPONENT"
GSR_TESTING_NAME = "GSR_TESTING_COMPONENT"

TERMINATE_MESSAGE_TYPE = "GSRTERMINATE"
INFO_MESSAGE_TYPE = "INFOMESSAGE"
UPDATE_MESSAGE_TYPE = "GSRUPDATE"
TERMINATE_ROUTING_MESSAGE_TYPE = "GSRTERMINATEROUTING"
ROUTING_COMPLETED_MESSAGE_TYPE = "GSRROUTINGCOMPLETED"

N_NODES = 10
ROUTER_UPDATE_PERIOD_SECS = 0.1
TOMBALA_DRAW_PERIOD_SECS = 0.2
TOMBALA_PLAYER_PERIOD_SECS = 0.1
TOMBALA_MAX_NUMBER = 90
TOMBALA_N_NUMBERS_IN_HAND = 9

ENABLE_NETWORK_LEVEL_LOGGING = False



#!/usr/bin/env python
import time
from ahc.Ahc import *
from itertools import chain

def startTreeAlgorithm(treeTopology):
  for node in treeTopology.nodes.values():
    node.startTreeAlgorithm()

class TreeNode(ComponentModel):

  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

  def on_message_from_bottom(self, eventobj: Event):
    for ch in self.unvisitedNeighbours:
      # print(f"EventSource: {eventobj.eventsource} Channel Conns: {chain(*ch.connectors.values())}")
      if eventobj.eventsource in chain(*ch.connectors.values()):
        channel = ch
        break
 
    self.unvisitedNeighbours.remove(channel)
    # self.parent = eventobj.eventsource
    if len(self.unvisitedNeighbours) == 0:
      self.decide()
    elif len(self.unvisitedNeighbours) == 1:
      self.parent = self.unvisitedNeighbours[0]
      print(f"{self.componentname}.{self.componentinstancenumber} sends message to {self.parent.componentname}.{self.parent.componentinstancenumber}")
      self.parent.trigger_event(Event(self, EventTypes.MFRT, eventobj.eventcontent))
    else:
      #do nothing
      pass

  def startTreeAlgorithm(self):
    self.unvisitedNeighbours = self.connectors[ConnectorTypes.DOWN]
    # print(f"{self.componentname}.{self.componentinstancenumber} Neighbours: {self.unvisitedNeighbours}")

    if len(self.connectors[ConnectorTypes.DOWN]) == 1:
      self.parent = self.unvisitedNeighbours[0]
      print(f"{self.componentname}.{self.componentinstancenumber} sends message to {self.parent.componentname}.{self.parent.componentinstancenumber}")
      self.send_down(Event(self, EventTypes.MFRT, None))
      self.unvisitedNeighbours = []

  def decide(self):
    print(f"{self.componentname}.{self.componentinstancenumber} decides.")
    print(f"End Time: {time.time()}")




#!/usr/bin/env python
import time
import random
from ahc.Ahc import *
from itertools import chain

#Please close prints if you are working with big graphs for better results.

def startEchoAlgorithm(echoTopology):
  # print(echoTopology.nodes)
  initiator = random.choice(list(echoTopology.nodes.values()))
  initiator.startEchoAlgorithm()

class EchoNode(ComponentModel):

  def __init__(self, componentname, componentid):
    super().__init__(componentname, componentid)
    self.isInitiator = False
    self.isFirstMessage = True

  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
    
  def on_message_from_bottom(self, eventobj: Event):
    if self.isFirstMessage:
      self.remainingNeighbours = self.connectors[ConnectorTypes.DOWN]      

    for ch in self.remainingNeighbours:
      # print(f"EventSource: {eventobj.eventsource} Channel Conns: {chain(*ch.connectors.values())}")
      if eventobj.eventsource in chain(*ch.connectors.values()):
        channel = ch
        break

    self.remainingNeighbours.remove(channel)

    if self.isFirstMessage:
      for ch in self.remainingNeighbours:
        print(f"{self.componentname}.{self.componentinstancenumber} sends message to {ch.componentname}.{ch.componentinstancenumber}")
        ch.trigger_event(Event(self, EventTypes.MFRT, eventobj.eventcontent))
      self.isFirstMessage = False
      self.parent = channel

    if len(self.remainingNeighbours) == 0:
      print(f"{self.componentname}.{self.componentinstancenumber} has received message from all neighbours.")
      if self.isInitiator:
        self.decide()

      else:
        print(f"{self.componentname}.{self.componentinstancenumber} sends message to parent {self.parent.componentname}.{self.parent.componentinstancenumber}")
        self.parent.trigger_event(Event(self, EventTypes.MFRT, eventobj.eventcontent))
      
    else:
      # print(f"{self.componentname}.{self.componentinstancenumber}: Channel {channel.componentinstancenumber} removed")
      #do nothing
      pass

  def send_down(self, event: Event):
    try:
      for p in self.connectors[ConnectorTypes.DOWN]:
        print(f"{self.componentname}.{self.componentinstancenumber} sends message to {p.componentname}.{p.componentinstancenumber}")
        p.trigger_event(event)
    except:
      pass

  def startEchoAlgorithm(self):
    self.isInitiator = True
    self.isFirstMessage = False
    self.remainingNeighbours = self.connectors[ConnectorTypes.DOWN]      
    print(f"{self.componentname}.{self.componentinstancenumber} is the initiator.")
    
    self.send_down(Event(self, EventTypes.MFRT, None))

  def decide(self):
    print(f"{self.componentname}.{self.componentinstancenumber} decides.")
    print(f"End Time: {time.time()}")



from enum import Enum
from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader
from ahc.Ahc import EventTypes
from collections import defaultdict

class SnapshotEventTypes(Enum):
    # Take snapshot event
    TS = "TS"

class SnapshotMessageTypes(Enum):
    GSU = "GSU"


class SnapshotComponentModel(ComponentModel):

    def __init__(self, componentname, componentinstancenumber,
                 num_worker_threads=1):
        super().__init__(componentname, componentinstancenumber,
                         num_worker_threads=num_worker_threads)
        self.state = None
        self.gsu_redirected_comps = set()
        self.recv_events = []
        self.chnls = set()
        self.init_snapshot = False
        self.eventhandlers[SnapshotEventTypes.TS] = self.take_snapshot

    def connect_me_to_component(self, name, component):
        raise Exception(f"Only channels are allowed for connection to"
                        " {self.__class__}")

    def on_connected_to_channel(self, name, channel):
        super().on_connected_to_channel(name, channel)
        self.chnls.add(channel.componentinstancenumber)

    def channel_of(self, eventobj: Event):
        from_chnl = eventobj.fromchannel
        if from_chnl is None:
            raise Exception(f"Received a message from a non-channel component")

        return from_chnl

    def on_pre_event(self, event):
        return self.recv_events.append(event)

    def msg_recv(self, event: Event):
        """Generic message received function"""
        pass

    def send_msg(self, event: Event):
        pass

    def send_gsu(self, local_state):
        gsu_msg = GenericMessage(
            GenericMessageHeader(SnapshotMessageTypes.GSU, None, None),
            local_state)
        self.send_msg(Event(self, EventTypes.MFRT, gsu_msg))

    def gsu_recv(self, state):
        # Redirect the GSU if we are not the source component of the snapshot
        if state.component_id not in self.gsu_redirected_comps:
            self.gsu_redirected_comps.add(state.component_id)
            self.send_gsu(state)

        self.on_gsu_recv(state)

    def on_gsu_recv(self, state):
        pass

    def on_take_snapshot(self):
        """Generic report snapshot"""
        pass

    def take_snapshot(self, eventobj: Event):
        self.init_snapshot = True
        self.on_take_snapshot()

    # When overridden call this function with 'super'
    def on_message_from_bottom(self, eventobj: Event):
        return self.msg_recv(eventobj)

    # When overridden call this function with 'super'
    def on_message_from_peer(self, eventobj: Event):
        return self.msg_recv(eventobj)

    # When overridden call this function with 'super'
    def on_message_from_top(self, eventobj: Event):
        return self.msg_recv(eventobj)

    def reset_state(self):
        self.state = None
        self.gsu_redirected_comps.clear()

class ChandyLamportMessageTypes(Enum):
    MARK = "MARK"


class ChandyLamportState:
    def __init__(self, component, state, chnl_states):
        self.component_id = component
        self.component_state = []
        for s in state:
            self.component_state.append(s)

        self.chnl_states = defaultdict(list)
        for c, s in chnl_states.items():
            self.chnl_states[c].append(s)

class ChandyLamportComponentModel(SnapshotComponentModel):
    """A ComponentModel that you can take a snapshot of using the
    Chandy-Lamport algorithm"""

    def __init__(self, componentname, componentinstancenumber,
                     num_worker_threads=1):
        super().__init__(componentname, componentinstancenumber,
                         num_worker_threads=num_worker_threads)
        self.global_state = dict()
        self.in_chnl_states = defaultdict(list)
        self.in_chnl_events = defaultdict(list)
        self.mark_recv_chnls = set()
        self.gsu_chnls = set()

    def on_gsu_recv(self, state: ChandyLamportState):
        if not self.init_snapshot:
            return

        report=f"State of component: {state.component_id}="
        report += ", ".join(str(e) for e in state.component_state)
        print(report)
        for chnl, events in state.chnl_states.items():
            chnl_rep = f"State of channel: {chnl}="
            chnl_rep += ", ".join(str(e) for e in events)
            print(chnl_rep)

    def send_msg(self, event: Event):
        self.send_down(event)

    def mark_send(self):
        # Record the state
        self.state = []
        for re in self.recv_events:
            self.state.append(re)

        # Broadcast the mark message
        mark_msg = GenericMessage(
            GenericMessageHeader(ChandyLamportMessageTypes.MARK, None, None),
            None)
        self.send_msg(Event(self, EventTypes.MFRT, mark_msg))

    def on_take_snapshot(self):
        """Initializes a global snapshot and a report will be printed out when
        complete"""
        self.mark_send()

    def mark_recv(self, from_chnl):
        if self.state is None:
            # First mark message, save component and channel state
            self.mark_send()
            self.in_chnl_states[from_chnl] = []
        else:
            # Consequent mark messages, save channel states
            for e in self.in_chnl_events[from_chnl]:
                self.in_chnl_states[e].append(e)

        self.mark_recv_chnls.add(from_chnl)
        if self.mark_recv_chnls == self.chnls:
            # Local snapshot completed, broadcast the local state
            local_state = ChandyLamportState(self.componentinstancenumber,
                                             self.state, self.in_chnl_states)
            gsu_msg = GenericMessage(
                GenericMessageHeader(SnapshotMessageTypes.GSU, None, None),
                local_state)
            self.send_msg(Event(self, EventTypes.MFRT, gsu_msg))
            self.gsu_recv(local_state)

    def msg_recv(self, event: Event):
        from_chnl = self.channel_of(event)
        # If received message is of type MARK or GSU; process them separately
        if type(contnt := event.eventcontent) == GenericMessage and\
           type(header := contnt.header) == GenericMessageHeader:
            if header.messagetype == ChandyLamportMessageTypes.MARK:
                self.mark_recv(from_chnl)
            elif header.messagetype == SnapshotMessageTypes.GSU:
                self.gsu_recv(contnt.payload)

            return event

        if self.state is None:
            return event

        # If the state is not recorded
        if from_chnl not in self.in_chnl_states:
            self.in_chnl_events[from_chnl].append(event)

        return event

    def reset_state(self):
        super().reset_state()
        self.in_chnl_states.clear()
        self.in_chnl_events.clear()
        self.mark_recv_chnls.clear()


class LaiYangState:
    def __init__(self, comp_id, comp_state, received, sent):
        self.component_id = comp_id
        self.component_state = []
        for cs in comp_state:
            self.component_state.append(cs)

        self.received = defaultdict(list)
        for chnl, r in received.items():
            self.received[chnl].append(r)

        self.sent = defaultdict(list)
        for chnl, s in sent.items():
            self.sent[chnl].append(s)

class LaiYangComponentModel(SnapshotComponentModel):
    def __init__(self, componentname, componentinstancenumber,
                 num_worker_threads=1):
        super().__init__(componentname, componentinstancenumber,
                         num_worker_threads=num_worker_threads)
        self.chnl_recv = defaultdict(list)
        self.chnl_sent = defaultdict(list)
        self.global_state = dict()
        self.sent_remaining = dict()
        self.recv_remaining = dict()

    def send_msg(self, event: Event):
        event.eventcontent = (event.eventcontent, self.state is not None)
        for c in self.chnls:
            self.chnl_sent[c].append(event)

        self.send_down(event)

    def handle_snapshot(self):
        # Take a snapshot
        self.state = LaiYangState(self.componentinstancenumber,
                                  self.recv_events, self.chnl_recv,
                                  self.chnl_sent)
        self.gsu_recv(self.state)

    def on_take_snapshot(self):
        self.handle_snapshot()

        # Broadcast a dummy message so that other components record
        # and broadcast their snapshots
        self.send_msg(Event(self, EventTypes.MFRT, "dummy"))

    def report_and_save_channel_state(self, channel, set_recv, set_sent):
        if not set_recv.issubset(set_sent):
            raise Exception("Not a consistent global state")

        chnl_state = list(set_sent - set_recv)
        self.global_state[channel] = chnl_state
        print(f"State of channel: {channel}=chnl_state")

    def on_gsu_recv(self, state: LaiYangState):
        if not self.init_snapshot:
            return
        # Report the snapshot if we are the source component of the snapshot
        self.global_state[state.component_id] = state.component_state
        report = f"State of component: {state.component_id}="
        report += ", ".join(str(e) for e in state.component_state)
        print(report)

        # Compute the messages in transit
        for chnl, recv in state.received:
            if chnl in self.sent_remaining:
                self.report_and_save_channel_state(
                    chnl, set(recv), set(self.sent_remaining[chnl]))
            else:
                self.recv_remaining[chnl] = recv

        for chnl, sent in state.sent:
            if chnl in self.recv_remaining:
                self.report_and_save_channel_state(
                    chnl, set(self.recv_remaining[chnl]), set(sent))
            else:
                self.sent_remaining[chnl] = sent

    def msg_recv(self, event: Event):
        content = event.eventcontent
        if type(content) is not tuple or len(content) != 2:
            raise Exception("Malformed message received by: "
                            "{self.unique_name()}")

        # Unpack the event content and modify the event with the actual content
        act_cntnt, post_snapshot = content
        event.eventcontent = act_cntnt

        # We are white and the message is post-snapshot
        if self.state is None and post_snapshot:
            self.handle_snapshot()

        from_chnl = self.channel_of(event)
        self.chnl_recv[from_chnl].append(event)

        # If not a GSU message return the modified event
        if type(act_cntnt) != GenericMessage or\
           type(header := act_cntnt.header) != GenericMessageHeader or\
               header.messagetype != SnapshotMessageTypes.GSU:
            return event

        self.gsu_recv(act_cntnt.payload)
        return event

    def reset_state(self):
        super().reset_state()
        self.global_state.clear()



from ahc.Routing.SSBR.HelperFunctions import SSBRRouteSearchMessage, SSBRUnicastMessage, messageParser
from ahc.Ahc import ComponentModel, Event, EventTypes

class ApplicationAndNetwork(ComponentModel):
    def __init__(self, componentname, componentid):
        super(ApplicationAndNetwork, self).__init__(componentname, componentid)
        self.componentid = componentid
        self.Response_Record = {}

    def on_init(self, eventobj: Event):
        print(f"{self.componentname} - #{self.componentid} is up.\n")
        #evt = Event(self, EventTypes.MFRT, messageGenerator(self))
        #self.send_down(evt)

    def on_message_from_bottom(self, eventobj: Event):
        messagePayload = eventobj.eventcontent.payload
        messageFrom = eventobj.eventcontent.header.messagefrom
        #print(f"{self.componentname}-{self.componentid} got a message from {messageFrom}. \n Message is {messagePayload}\n")

        if eventobj.eventcontent.header.messagetype == "UNICASTDATA":
            if eventobj.eventcontent.payload == "benchmark test message":
                evt = Event(self, EventTypes.MFRT, SSBRUnicastMessage(self, eventobj.eventcontent.header.messageto, eventobj.eventcontent.payload))
                self.send_down(evt) 
            else:
                print(f"Message from {eventobj.eventcontent.header.messagefrom} is delivered to {self.componentname}-{self.componentid}. Message is - {eventobj.eventcontent.payload}")
                if eventobj.eventcontent.header.messageto != str(self.componentname) + "-" + str(self.componentid):
                    evt = Event(self, EventTypes.MFRT, messageParser(self, eventobj))
                    self.send_down(evt)


    def send_test_message(self):
        evt = Event(self, EventTypes.MFRT, SSBRRouteSearchMessage(self))
        self.send_down(evt) 

    def send_SSBR_unicast_message(self, target, message=""):
        evt = Event(self, EventTypes.MFRT, SSBRUnicastMessage(self, target, message))
        self.send_down(evt) 






import random as rand
from types import new_class
from typing import List, Set, Tuple
from uuid import uuid4
import time

from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes,Topology

from ahc.Routing.ARA.ARAConfiguration import ARAConfiguration
from ahc.Routing.ARA.ARAEventTypes import ARAEventTypes
from ahc.Routing.ARA.ARARoutingEntry import ARARoutingEntry

"""
  implementation of "ARA-the ant-colony based routing algorithm for MANETs"
  [online] https://www.researchgate.net/publication/3971279_ARA-the_ant-colony_based_routing_algorithm_for_MANETs
"""

class RoutingARAComponent(ComponentModel):
  def __init__(self, componentname, componentid):
    super().__init__(componentname, componentid)

    # Information for ARA algorithm
    self.componentId = componentid
    self.neighbors = [] # the list of neighbors (ids) connected to the node
    self.entryTable: List[ARARoutingEntry] = [] # stores the routing table information
    self.visitedAnts: Set = set() # use set for unique tuples
    self.sequenceNums = []
    self.msgCache = [] # to cache the message and send them when route is found


  def on_init(self, eventobj: Event):
      super(RoutingARAComponent, self).on_init(eventobj)
      self.neighbors = Topology().get_neighbors(self.componentinstancenumber)

  def on_message_from_bottom(self, eventobj: Event):
    if (int(eventobj.eventcontent.header.interfaceid.split("-")[1]) == self.componentId):
      self.ARA(eventobj)

  def ARA(self, eventobj: Event):
    message_target = eventobj.eventcontent.header.messageto.split("-")[0]
    message_target_id = eventobj.eventcontent.header.messageto.split("-")[1]

    if message_target == "RoutingARAComponent":
      # for each message we should evaporate the pheromone value
      for entry in self.entryTable:
        entry.evaporatePheromone()

      message_header = eventobj.eventcontent.header    
      if (message_header.messagetype == ARAEventTypes.REGULAR):
        if (not self.pass_regular_packet(eventobj)):
          self.msgCache.append(eventobj) # cache msg for route found
          forward_ant_message = self.make_ant_packet(self.componentId, message_target_id, ARAEventTypes.FANT)
          self.send_to_neighbors(forward_ant_message)
      elif (message_header.messagetype == ARAEventTypes.FANT):
        self.pass_discovery_ant(eventobj)
      elif (message_header.messagetype == ARAEventTypes.BANT):
        self.pass_discovery_ant(eventobj)
      elif (message_header.messagetype == ARAEventTypes.DUPLICATE_ERROR):
        self.deactivate_loop_route(eventobj)
      elif (message_header.messagetype == ARAEventTypes.ROUTE_ERROR):
        # this case is for dynamic networks, hence we can ignore it
        return
      else :
        print("ERROR Unknown Message Type", message_header.messagetype)

  def pass_regular_packet(self, eventobj: Event) -> bool:
    # check if we are the destination of the packet
    messageto = eventobj.eventcontent.header.messageto.split("-")[1]
    if (self.componentId != int(messageto)):
      entries = self.find_destination_entries(int(messageto))
      
      if(len(entries) == 0):
        return False
      else:
        if (eventobj.eventcontent.header.sequencenumber in self.sequenceNums):
          # means that we have seen this data package and we are in a loop situation for regular packet
          self.pass_loop_packet(eventobj)
        else: 
          self.sequenceNums.append(eventobj.eventcontent.header.sequencenumber)
          next_hop = self.find_next_hop(entries)
          self.send_to_one_neighbor(eventobj.eventcontent, next_hop.nextHopAddress)
          next_hop.increasePheromone()
    else:
      if (int(eventobj.eventcontent.header.messagefrom.split("-")[1]) != self.componentId):
        if (eventobj.eventcontent.header.sequencenumber not in self.sequenceNums):
          self.sequenceNums.append(eventobj.eventcontent.header.sequencenumber)

    return True

  def pass_discovery_ant(self, eventobj: Event):
    previous_address = eventobj.eventcontent.header.interfaceid.split("-")[0]
    messagefrom = eventobj.eventcontent.header.messagefrom.split("-")[1]
    messageto = eventobj.eventcontent.header.messageto.split("-")[1]

    entry = self.get_entry_for_destination(int(messagefrom), int(previous_address))
    
    if (entry is None and int(messagefrom) != self.componentId):
      # make new entry and push it to entries list
      new_entry = ARARoutingEntry(int(messagefrom), int(previous_address))
      new_entry.increasePheromone()
      self.entryTable.append(new_entry)

    # try to push new tuple to visitedAnts and continue if not visited before
    if(self.do_add_to_set(s = self.visitedAnts, added=(eventobj.eventcontent.header.sequencenumber, int(previous_address)))):
      if (int(messageto) == self.componentId):
        if (eventobj.eventcontent.header.messagetype == ARAEventTypes.FANT):
          backward_ant_msg = self.make_ant_packet(
            source=self.componentId,
            destination= int(messagefrom),
            msg_type = ARAEventTypes.BANT)
          self.send_to_neighbors(backward_ant_msg)
        else:
          # means that bant is acquired at the right place, hence find the first message and send it with ARA
          cached_event = self.find_cached_event(eventobj)
          if cached_event is not None: 
            # deep copy the element since we are going to delete it
            cp_cached_event = self.copy_event(cached_event)
            self.msgCache.remove(cached_event)
            self.ARA(cp_cached_event) 
      else:
        # send to all neightbors except previous address
        self.send_to_neighbors(eventobj.eventcontent, previous_address)
    else:
      # ignore the packet
      return

  def pass_loop_packet(self, eventobj: Event):
    previous_sender = int(eventobj.eventcontent.header.interfaceid.split("-")[0])
    message_header = GenericMessageHeader(
      messagetype = ARAEventTypes.DUPLICATE_ERROR,
      messagefrom = eventobj.eventcontent.header.messageto,
      messageto = "RoutingARAComponent-"+str(self.componentId),
      nexthop = previous_sender,
      sequencenumber = uuid4()) # unique sequence number to identify ants
    payload = eventobj # send previous event object in payload to continue searching
    message = GenericMessage(message_header, payload)
    self.send_to_one_neighbor(message, previous_sender)

  def deactivate_loop_route(self, eventobj: Event):
    messageto = eventobj.eventcontent.header.messageto.split("-")[1]
    messagefrom = eventobj.eventcontent.header.messagefrom.split("-")[1]
    target = eventobj.eventcontent.header.interfaceid.split("-")[1]
    previous_msg = eventobj.eventcontent.payload # previous msg is send back in the payload

    if (int(target) == self.componentId):
      # delete the looping route
      for entry in self.entryTable:
        if (entry.destination == int(messagefrom) 
            and 
            entry.nextHopAddress == int(messageto)):
          self.entryTable.remove(entry)

      # handle to previous packet but change its id so fresh start
      new_msg_header = GenericMessageHeader(
        messagetype = previous_msg.eventcontent.header.messagetype,
        messagefrom = previous_msg.eventcontent.header.messagefrom,
        messageto = previous_msg.eventcontent.header.messageto,
        sequencenumber = uuid4())
      payload = ""
      msg = GenericMessage(new_msg_header, payload) 
      self.ARA(Event(self, EventTypes.MFRT, msg))

    else:
      print("ERROR, received DUPLICATE_ERROR was not addressed to me: ", self.componentId)
    
  def do_add_to_set(self, s: Set, added: Tuple) -> bool:
    # checks if adding operation is successfull to set
    l = len(s)
    s.add(added)
    return len(s) != l


  def find_destination_entries(self, destination_adress) -> List[ARARoutingEntry]:
    entries: List[ARARoutingEntry] = []
    for entry in self.entryTable:
      if (entry.destination == destination_adress):
        entries.append(entry)
    return entries

  def find_next_hop(self, entries: List[ARARoutingEntry]) -> ARARoutingEntry:
    total_pheromone = 0.0
    for entry in entries:
      total_pheromone = total_pheromone + entry.pheromone
    
    randomVar = (rand.randint(0, ARAConfiguration.RAND_MAX) / ARAConfiguration.RAND_MAX) * total_pheromone 

    choosenHop = -1
    while (randomVar >= 0.0):
      choosenHop = choosenHop + 1
      if (len(entries) == choosenHop):
        return entries[-1]
      randomVar = randomVar - entries[choosenHop].pheromone

    return entries[choosenHop] 

  def make_ant_packet(self, source, destination, msg_type: ARAEventTypes) -> GenericMessage:
    message_header = GenericMessageHeader(
      messagetype = msg_type,
      messagefrom = "RoutingARAComponent-"+str(source),
      messageto = "RoutingARAComponent-"+str(destination),
      nexthop = 0, # total hop count as header
      sequencenumber = uuid4()) # unique sequence number to identify ants
    payload = "" # empty payload for ants
    message = GenericMessage(message_header, payload)
    return message

  def get_entry_for_destination(self, destination, hop_adress) -> ARARoutingEntry:
    for entry in self.entryTable:
      if (entry.destination == destination and entry.nextHopAddress == hop_adress):
        return entry
    return None

  def send_to_one_neighbor(self, packet: GenericMessage, neighbor):
    if (neighbor not in self.neighbors):
      print("ERROR, neighbor ", neighbor, " is not a neighbor of ", self.componentId)
    else:
      new_packet = self.create_new_packet_with_interface(packet, self.componentId, neighbor)    

      self.send_down(Event(self, EventTypes.MFRT, new_packet))

  def send_to_neighbors(self, packet: GenericMessage, exceptions=[]):
    for n in self.neighbors:
      if (len(exceptions) == 0):
        # create new packet with interface
        new_packet = self.create_new_packet_with_interface(packet, self.componentId, n)
        self.send_down(Event(self, EventTypes.MFRT, new_packet))

      else:
        for e in exceptions:
          if (n != int(e)):
            # set interface id for channels
            new_packet = self.create_new_packet_with_interface(packet, self.componentId, n)
            self.send_down(Event(self, EventTypes.MFRT, new_packet))

  def create_new_packet_with_interface(self, packet: GenericMessage, from_node, to_node) -> GenericMessage:
    message_header = GenericMessageHeader(
      messagetype=packet.header.messagetype,
      messagefrom=packet.header.messagefrom,
      messageto=packet.header.messageto,
      nexthop=to_node,
      interfaceid=str(from_node)+"-"+str(to_node), 
      sequencenumber=packet.header.sequencenumber)
    return GenericMessage(message_header, packet.payload)

  def find_cached_event(self, eventobj: Event) -> Event:
    messageto = eventobj.eventcontent.header.messageto.split("-")[1]
    messagefrom = eventobj.eventcontent.header.messagefrom.split("-")[1]

    # reverse the addresses since this func is called fro BANT event
    for e in self.msgCache:
      header = e.eventcontent.header
      if (header.messageto.split("-")[1] == messagefrom
          and
          header.messagefrom.split("-")[1] == messageto):
        return e
    return None

  def copy_event(self, eventobj: Event) -> Event:
    new_header = GenericMessageHeader(
      messagetype=eventobj.eventcontent.header.messagetype,
      messagefrom=eventobj.eventcontent.header.messagefrom,
      messageto=eventobj.eventcontent.header.messageto,
      nexthop=eventobj.eventcontent.header.nexthop,
      interfaceid=eventobj.eventcontent.header.interfaceid,
      sequencenumber=eventobj.eventcontent.header.sequencenumber
    )
    new_payload = eventobj.eventcontent.payload
    new_msg = GenericMessage(new_header, new_payload)
    new_event = Event(eventobj.eventsource, eventobj.event, new_msg, eventobj.fromchannel, eventobj.eventid)
    return new_event

  def make_interface_id(self, from_node, to_node):
    if int(self.from_node) > int(to_node):
        interfaceid = str(to_node)+"-"+str(from_node)
        return interfaceid
    else:
        interfaceid=str(from_node)+"-"+str(to_node)
        return interfaceid


import random
import networkx as nx
from threading import Thread, Event
from datetime import datetime

from ahc.Ahc import singleton


def random_undirected_graph(n):
    tries = 1000
    for i in range(tries):
        g = nx.gnp_random_graph(n, 0.2)
        if nx.is_connected(g):
            return g
    raise nx.NetworkXError("Maximum number of tries exceeded")


def random_directed_graph(n):
    g = random_undirected_graph(n).to_directed()
    # we can also sprinkle some directed edges
    # gd = nx.gnp_random_graph(n, 0.01, None, True)
    # for (u, v) in gd.edges:
    #    g.add_edge(u, v)
    return g


def random_layered_graph(n):
    def make_layer(mp):
        g = random_undirected_graph(len(mp))
        gs = nx.Graph()
        for (su, sv) in g.edges:
            gs.add_edge(mp[su], mp[sv])
        return gs
    i = list(range(n))
    random.shuffle(i)
    #  0- 50 => {1}
    # 50- 80 => {1,2}
    # 80- 90 =>   {2,3}
    # 90-100 =>     {3}
    sep = [make_layer(i[:int(.8*n)]),
           make_layer(i[int(.5*n):int(.9*n)]),
           make_layer(i[int(.7*n):])]
    g = nx.Graph()
    node_layers = {}
    for layer in range(len(sep)):
        for (u, v) in sep[layer].edges:
            g.add_edge(u, v)
            if u not in node_layers:
                node_layers[u] = set()
            node_layers[u].add(layer)
            if v not in node_layers:
                node_layers[v] = set()
            node_layers[v].add(layer)
    for k, v in node_layers.items():
        g.add_node(k, conf={'layers': v})
    return g, sep

def keys_to_set(m):
    s = set()
    for n in m.keys():
        s.add(n)
    return s


@singleton
class Tracing:
    nodes = set()
    node_address_to_id = {}
    node_edges = {}
    node_edges_each_step = []

    def register_node(self, address, node_id):
        self.nodes.add(node_id)
        self.node_address_to_id[address] = node_id

    def update_edges(self, edges):
        current_time = datetime.now()
        ne = {}
        for k, v in self.node_edges.items():
            ne[k] = v
        for (from_addr, to_addr) in edges:
            from_id = self.node_address_to_id[from_addr]
            to_id = self.node_address_to_id[to_addr]
            attr = {'received_at': current_time}
            ne[(from_id, to_id)] = attr
        if keys_to_set(self.node_edges) != keys_to_set(ne):
            self.node_edges_each_step.append(ne)
        self.node_edges = ne

    def to_graph(self):
        edges = []
        for (f, t), a in self.node_edges.items():
            edges.append((f, t, a))
        graph = nx.Graph()
        graph.add_nodes_from(self.nodes)
        graph.add_edges_from(edges)
        return graph

    def step_to_graph(self, step):
        edges = []
        if step >= len(self.node_edges_each_step):
            step = len(self.node_edges_each_step) - 1
        for (f, t), a in self.node_edges_each_step[step].items():
            edges.append((f, t, a))
        graph = nx.Graph()
        graph.add_nodes_from(self.nodes)
        graph.add_edges_from(edges)
        return graph


@singleton
class RepeatDeltaTimer(Thread):
    def __init__(self):
        Thread.__init__(self)
        self.interval = 1.0
        self.finished = Event()
        self.functions = []

    def set_interval(self, interval):
        self.interval = interval

    def register_function(self, fn):
        self.functions.append(fn)

    def cancel(self):
        self.finished.set()

    def run(self):
        prev_time = datetime.now()
        while not self.finished.wait(self.interval):
            current_time = datetime.now()
            delta = current_time - prev_time
            prev_time = current_time
            for f in self.functions:
                f(current_time, delta.total_seconds())



import os
import sys
import random

import numpy as np
import matplotlib.pyplot as plt
import time as t

sys.path.insert(0, os.getcwd())

from ahc.Ahc import ComponentRegistry, Topology
from ahc.Channels.Channels import  P2PFIFOPerfectChannel
import networkx as nx
import matplotlib.pyplot as plt
from ahc.Routing.SSBR.SSBRNode import SSBRNode
from ahc.Routing.SSBR.HelperFunctions import buildRoutingTable, findStrongConnectedLinksForSingleNode, printSSTInfo, constructStrongRoute, resetRoutingState, benchmarkTest

NODE_COUNT = input("Enter # of nodes: ")
NODE_COUNT = int(NODE_COUNT)

def draw_random_graph(n):
    k = True
    while k == True:
        k = False
        g_random = nx.gnp_random_graph(n, 0.45)
        if not nx.is_connected(g_random):
            k = True
    return g_random

# Generating random graph
graph =  draw_random_graph(NODE_COUNT)

# Changing weights
for (u,v,w) in graph.edges(data=True):

    w['weight'] = round(random.uniform(0.01,0.99), 2)

# Drawing the graph
topology = Topology()
topology.construct_from_graph(graph, SSBRNode, P2PFIFOPerfectChannel)

topology.start()
threshold = input("Enter threshold value for signal strength (between 0 an 1):\n")
threshold = float(threshold)

# Menu
print("\n 1. Show graph. \n 2. Print information. \n 3. Construct strongly connected route for a node. \n 4. Send unicast data between nodes. \n 5. Benchmark test")

menuItem = input("Enter a value to proceed:\n")
menuItem = int(menuItem)
SSBRForwardingTable = []

while(menuItem):

    if menuItem == 1:
        pos = nx.spring_layout(graph)
        labels = nx.get_edge_attributes(graph,'weight')
        #print(labels)
        nx.draw(graph, pos, with_labels=True)
        nx.draw_networkx_edge_labels(graph, pos, edge_labels=labels)

        plt.show(block=False)

    elif menuItem == 2:
        printSSTInfo(NODE_COUNT)

    elif menuItem == 3:
        source = int(input("Enter source node id:\n"))
        target = int(input("Enter target node id:\n"))
        resetRoutingState(NODE_COUNT)
        labels = nx.get_edge_attributes(graph,'weight')
        findStrongConnectedLinksForSingleNode(labels, threshold, NODE_COUNT)
        buildRoutingTable(source, target)
        print(constructStrongRoute(graph, source, target))
        
    elif menuItem == 4:
        source = int(input("Enter source node id:\n"))
        target = int(input("Enter target node id:\n"))
        labels = nx.get_edge_attributes(graph,'weight')
        resetRoutingState(NODE_COUNT)
        findStrongConnectedLinksForSingleNode(labels, threshold, NODE_COUNT)
        buildRoutingTable(source, target)
        SSBRForwardingTable = constructStrongRoute(graph, source,target)

        if len(SSBRForwardingTable) >= 1:
            sourceNode = ComponentRegistry().get_component_by_key("ApplicationAndNetwork", source)
            sourceNode.send_SSBR_unicast_message(target)
            SSBRForwardingTable=[]
        else:
            print(f"No possible route between #{source}-#{target}")

    elif menuItem == 5:
        rows, cols = (10, 10)
        nodeNumber = [[0 for i in range(cols)] for j in range(rows)]
        time = [[0 for i in range(cols)] for j in range(rows)]
        controlTime = [[0 for i in range(cols)] for j in range(rows)]
        dataTime = [[0 for i in range(cols)] for j in range(rows)]
        for i in range (0,10):
            for x in range (3,13):
                NODE_COUNT = x
                graph =  draw_random_graph(NODE_COUNT)
                for (u,v,w) in graph.edges(data=True):
                    w['weight'] = round(random.uniform(0.01,0.99), 2)
                topology = Topology()
                topology.construct_from_graph(graph, SSBRNode, P2PFIFOPerfectChannel)
                topology.start()
                source = 0
                target = x-1
                labels = nx.get_edge_attributes(graph,'weight')
                start_time = t.time()
                resetRoutingState(NODE_COUNT)
                findStrongConnectedLinksForSingleNode(labels, threshold, NODE_COUNT)
                controlTime[i][x-3] = (t.time()-start_time)
                start_time2 = t.time()
                benchmarkTest(NODE_COUNT)
                nodeNumber[i][x-3] = x
                dataTime[i][x-3] = (t.time()-start_time2)
                time[i][x-3] = (t.time()-start_time)
                for y in range (0,x):
                    sourceNode = ComponentRegistry().get_component_by_key("SSBRNode", source)
                    sourceNode.terminate()
                    sourceNode = ComponentRegistry().get_component_by_key("ApplicationAndNetwork", source)
                    sourceNode.terminate()
                    sourceNode = ComponentRegistry().get_component_by_key("FP", source)
                    sourceNode.terminate()
                    sourceNode = ComponentRegistry().get_component_by_key("DRP", source)
                    sourceNode.terminate()
                    sourceNode = ComponentRegistry().get_component_by_key("NetworkInterface", source)
                    sourceNode.terminate()

        resTime = [0]*10
        resNode = [0]*10
        resControlTime = [0]*10
        resDataTime = [0]*10

        for i in range (0,10):
            for k in range (0,10):
                resTime[i] += time[k][i]
                resNode[i] += nodeNumber[k][i]
                resControlTime[i] += controlTime[k][i]
                resDataTime[i] += dataTime[k][i]

        for i in range (0,10):
            resTime[i] /= 10
            resNode[i] /= 10
            resControlTime[i] /= 10
            resDataTime[i] /= 10

        plt.plot(resNode, resTime, label = "Overall Time")
        plt.plot(resNode, resControlTime, label = "Control Plane Time")
        plt.plot(resNode, resDataTime, label = "Data Plane Time")
        plt.grid(color='b', linestyle='-', linewidth=0.2)
        plt.legend()
        plt.title('Node number vs Time')
        plt.xlabel('Node Number')
        plt.ylabel('Time')
        plt.show()

    menuItem = input("Enter a new value to proceed:\n")
    print("\n 1. Show graph. \n 2. Print information. \n 3. Construct strongly connected route for a node. \n 4. Send unicast data between nodes. \n 5. Benchmark test")
    menuItem = int(menuItem)


while True: pass



from enum import Enum

from ahc.Ahc import ComponentModel, MessageDestinationIdentifiers, Event, GenericMessageHeader, GenericMessagePayload, \
  GenericMessage, EventTypes

# define your own message types
class LinkLayerMessageTypes(Enum):
  LINKMSG = "LINKMSG"

# define your own message header structure
class LinkLayerMessageHeader(GenericMessageHeader):
  pass

# define your own message payload structure
class LinkLayerMessagePayload(GenericMessagePayload):
  pass

class LinkLayer(ComponentModel):
  def on_init(self, eventobj: Event):
    # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
    pass

  def on_message_from_top(self, eventobj: Event):
    abovehdr = eventobj.eventcontent.header
    if abovehdr.messageto == MessageDestinationIdentifiers.NETWORKLAYERBROADCAST:
      hdr = LinkLayerMessageHeader(LinkLayerMessageTypes.LINKMSG, self.componentinstancenumber,
                                   MessageDestinationIdentifiers.LINKLAYERBROADCAST,nexthop=MessageDestinationIdentifiers.LINKLAYERBROADCAST)
    else:
      #if we do not broadcast, use nexthop to determine interfaceid and set hdr.interfaceid
      myinterfaceid = str(self.componentinstancenumber) + "-" + str(abovehdr.nexthop)
      hdr = LinkLayerMessageHeader(LinkLayerMessageTypes.LINKMSG, self.componentinstancenumber,
                                   abovehdr.nexthop, nexthop=abovehdr.nexthop, interfaceid=myinterfaceid)

    payload = eventobj.eventcontent
    msg = GenericMessage(hdr, payload)
    self.send_down(Event(self, EventTypes.MFRT, msg))

  def on_message_from_bottom(self, eventobj: Event):
    msg = eventobj.eventcontent
    hdr = msg.header
    payload = msg.payload
    if hdr.messageto == self.componentinstancenumber or hdr.messageto == MessageDestinationIdentifiers.LINKLAYERBROADCAST:
      self.send_up(Event(self, EventTypes.MFRB, payload,
                         eventobj.fromchannel))  # doing decapsulation by just sending the payload
    else:
      #      print(f"I am {self.componentinstancenumber} and dropping the {hdr.messagetype} message to {hdr.messageto}")
      # Physical layer is a broadcast medium, and hence will accept all messages. The link layer will drop those messages that are not for me
      pass

#  def __init__(self, componentname, componentinstancenumber):
#   super().__init__(componentname, componentinstancenumber)
#   If you have to extend the model add events here
